import glob
import os

from nipype.pipeline.engine import Workflow, Node
from nipype.interfaces.utility import IdentityInterface, Function
from nipype.interfaces.io import DataSink

from nipype.interfaces.ants.registration import RegistrationSynQuick
from nipype.interfaces.ants.resampling import ApplyTransforms
from qsmxt.interfaces import nipype_interface_fastsurfer as fastsurfer
from qsmxt.interfaces import nipype_interface_mgz2nii as mgz2nii
from qsmxt.interfaces import nipype_interface_analyse as analyse
from qsmxt.interfaces import nipype_interface_romeo as romeo
from qsmxt.interfaces import nipype_interface_tgv_qsm_jl as tgvjl
from qsmxt.interfaces import nipype_interface_qsmjl as qsmjl
from qsmxt.interfaces import nipype_interface_nextqsm as nextqsm
from qsmxt.interfaces import nipype_interface_laplacian_unwrapping as laplacian
from qsmxt.interfaces import nipype_interface_processphase as processphase
from qsmxt.interfaces import nipype_interface_axialsampling as sampling
from qsmxt.interfaces import nipype_interface_t2star_r2star as t2s_r2s
from qsmxt.interfaces import nipype_interface_clearswi as swi
from qsmxt.interfaces import nipype_interface_nonzeroaverage as nonzeroaverage
from qsmxt.interfaces import nipype_interface_twopass as twopass
from qsmxt.interfaces import nipype_interface_resample_like as resample_like
from qsmxt.interfaces import nipype_interface_qsm_referencing as qsm_referencing
from qsmxt.interfaces import nipype_interface_nii2dcm as nii2dcm
from qsmxt.interfaces import nipype_interface_copyfile as copyfile

from qsmxt.scripts.logger import LogLevel, make_logger
from qsmxt.scripts.qsmxt_functions import gen_plugin_args, create_node
from qsmxt.workflows.masking import masking_workflow

import numpy as np
import nibabel as nib


def insert_before(wf, target_node_name, new_node, target_attribute):
    """
    Inserts a new node before a specified target node in a Nipype workflow, updating existing connections.

    :param wf: The workflow object.
    :param target_node_name: The name of the target node.
    :param new_node: The new Node object to insert.
    :param target_attribute: The target attribute in the target node to which connections are made.
    """
    target_node = wf.get_node(target_node_name)

    # Find all source nodes connected to the target attribute of the target node
    source_edges = [(u, v, d) for u, v, d in wf._graph.edges(data=True) if v._name == target_node_name and d['connect'][0][1] == target_attribute]

    # Disconnect the source nodes from the target node
    for u, v, d in source_edges:
        source_attribute = d['connect'][0][0]
        wf.disconnect([(u, v, [(source_attribute, target_attribute)])])

    # Add the new node to the workflow
    wf.add_nodes([new_node])

    # Connect the source nodes to the new node and the new node to the target node
    for u, _, _ in source_edges:
        source_attribute = d['connect'][0][0]
        wf.connect([(u, new_node, [(source_attribute, target_attribute)])])
        wf.connect([(new_node, target_node, [(target_attribute, target_attribute)])])

def get_node(wf, node_name):
    for node in wf._get_all_nodes():
        if node_name in node._name:
            return node
    return None

def get_preceding_node_and_attribute(wf, target_node_name, target_attribute):
    """
    Retrieves the input node and its output attribute name connected to a specific attribute 
    of a target node in a Nipype workflow.

    :param wf: The workflow object.
    :param target_node_name: The name of the target node.
    :param target_attribute: The target attribute in the target node.
    :return: A tuple (input node, output attribute name) if found, otherwise (None, None).
    """
    # Search for an edge where the target node and attribute match
    for u, v, d in wf._graph.edges(data=True):
        for source_attribute, target_attr in d['connect']:
            if v._name == target_node_name and target_attr == target_attribute:
                return u, source_attribute

    return None, None

def get_matching_files(bids_dir, subject, dtype="anat", suffixes=[], ext="nii*", session=None, space=None, run=None, part=None, acq=None):
    pattern = os.path.join(bids_dir, subject)
    if session:
        pattern = os.path.join(pattern, session)
    pattern = os.path.join(pattern, dtype) + os.path.sep
    if space:
        pattern += f"*space-{space}*"
    if acq:
        pattern += f"*acq-{acq}*"
    if run:
        pattern += f"*run-{run}*"
    if part:
        pattern += f"*part-{part}*"
    dir, fname = os.path.split(pattern)
    if suffixes:
        if fname:
            matching_files = [glob.glob(f"{pattern}_{suffix}.{ext}") for suffix in suffixes]
        else:
            matching_files = [glob.glob(os.path.join(dir, f"*{suffix}.{ext}")) for suffix in suffixes]
    else:
        matching_files = [glob.glob(f"{pattern}.{ext}")]
    return sorted([item for sublist in matching_files for item in sublist])

def init_qsm_workflow(run_args, subject, session=None, acq=None, run=None):
    logger = make_logger('main')
    run_id = f"{subject}" + (f".{session}" if session else "") + (f".acq-{acq}" if acq else "") + (f".run-{run}" if run else "")
    logger.log(LogLevel.INFO.value, f"Creating QSMxT workflow for {run_id}...")

    # get relevant files from this run
    t1w_files = get_matching_files(run_args.bids_dir, subject=subject, dtype="anat", suffixes=["T1w"], ext="nii*", session=session, run=None, part=None, acq=None)
    phase_files = get_matching_files(run_args.bids_dir, subject=subject, dtype="anat", suffixes=[], session=session, run=run, part="phase", acq=acq)[:run_args.num_echoes]
    magnitude_files = get_matching_files(run_args.bids_dir, subject=subject, dtype="anat", suffixes=[], session=session, run=run, part="mag", acq=acq)[:run_args.num_echoes]
    phase_params_files = [path.replace('.nii.gz', '.nii').replace('.nii', '.json') for path in phase_files]
    mag_params_files = [path.replace('.nii.gz', '.nii').replace('.nii', '.json') for path in magnitude_files]
    params_files = phase_params_files if len(phase_params_files) else mag_params_files
    mask_files = [
        mask_file for mask_file in get_matching_files(os.path.join(run_args.bids_dir, "derivatives", run_args.existing_masks_pipeline), subject=subject, dtype="anat", suffixes=["mask"], session=session, run=None, part=None, acq=None)[:run_args.num_echoes]
        if ('_space-orig' in mask_file or '_space-' not in mask_file)
        and ('_label-brain' in mask_file or '_label-' not in mask_file)
        and ('qsmxt-workflow' not in mask_file)
    ]
    qsm_files = [
        qsm_file for qsm_file in get_matching_files(os.path.join(run_args.bids_dir, "derivatives", run_args.existing_qsm_pipeline), subject=subject, dtype="anat", suffixes=["Chimap"], session=session, run=None, part=None, acq=None)
        if ('qsmxt-workflow' not in qsm_file)
    ]
    seg_files = [
        seg_file for seg_file in get_matching_files(os.path.join(run_args.bids_dir, "derivatives", run_args.existing_segmentation_pipeline), subject=subject, dtype="anat", suffixes=["dseg"], session=session, space="qsm", run=None, part=None, acq=None)
        if ('qsmxt-workflow' not in seg_file)
    ]

    # handle any errors related to files and adjust any settings if needed
    if run_args.do_segmentation and not t1w_files:
        logger.log(LogLevel.WARNING.value, f"{run_id}: Skipping segmentation - no T1w files found!")
        run_args.do_segmentation = False
    if run_args.do_segmentation and not magnitude_files:
        logger.log(LogLevel.WARNING.value, f"{run_id}: Skipping segmentation - no GRE magnitude files found to register T1w segmentations to!")
        run_args.do_segmentation = False
    if run_args.do_segmentation and len(t1w_files) > 1:
        logger.log(LogLevel.WARNING.value, f"{run_id}: Using {t1w_files[0]} for segmentation - multiple T1w files found!")
    if run_args.do_qsm and not phase_files:
        logger.log(LogLevel.WARNING.value, f"{run_id}: Skipping QSM - No phase files found!")
        run_args.do_qsm = False
        run_args.do_swi = False
    if len(phase_files) != len(phase_params_files) and any([run_args.do_qsm, run_args.do_swi]):
        logger.log(LogLevel.ERROR.value, f"{run_id}: An unequal number of JSON and phase files are present - QSM and SWI are not possible!")
        run_args.do_qsm = False
        run_args.do_swi = False
    if len(magnitude_files) != len(mag_params_files) and any([run_args.do_swi, run_args.do_r2starmap, run_args.do_t2starmap, run_args.do_segmentation]):
        logger.log(LogLevel.ERROR.value, f"{run_id}: An unequal number of JSON and magnitude files are present - SWI, R2* mapping, T2* mapping, and segmentation are not possible!")
        run_args.do_swi = False
        run_args.do_r2starmap = False
        run_args.do_t2starmap = False
        run_args.do_segmentation = False
    if len(phase_files) == 1 and run_args.combine_phase:
        run_args.combine_phase = False
    if run_args.do_qsm and run_args.use_existing_masks:
        if not mask_files:
            logger.log(LogLevel.WARNING.value, f"{run_id}: --use_existing_masks specified but no masks found in {run_args.use_existing_masks} derivatives. Reverting to {run_args.masking_algorithm} masking algorithm.")
            run_args.use_existing_masks = False
        else:
            if len(mask_files) > 1:
                if run_args.combine_phase:
                    logger.log(LogLevel.WARNING.value, f"{run_id}: --combine_phase specified but multiple masks found with --use_existing_masks. Using the first mask only.")
                    mask_files = [mask_files[0]]
                elif len(mask_files) != len(phase_files):
                    logger.log(LogLevel.WARNING.value, f"{run_id}: --use_existing_masks specified but unequal number of mask and phase files present. Using the first mask only.")
                    mask_files = [mask_files[0]]
            if mask_files:
                run_args.inhomogeneity_correction = False
                run_args.two_pass = False
                run_args.add_bet = False
    if not magnitude_files:
        if run_args.do_r2starmap:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot compute R2* - no magnitude files found.")
            run_args.do_r2starmap = False
        if run_args.do_t2starmap:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot compute T2* - no magnitude files found.")
            run_args.do_t2starmap = False
        if run_args.do_swi:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot compute SWI - no magnitude files found.")
            run_args.do_swi = False
        if run_args.do_qsm:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot resample axially - no magnitude files were found (expect poor results only from oblique acquisitions)!")
            if run_args.masking_input == 'magnitude':
                logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot use magnitude-based masking - no magnitude files found!")
                run_args.masking_input = 'phase'
                run_args.masking_algorithm = 'threshold'
                run_args.inhomogeneity_correction = False
                run_args.add_bet = False
            if run_args.add_bet:
                logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot use --add_bet - no magnitude files found!")
                run_args.add_bet = False
    elif len(magnitude_files) != len(phase_files) and run_args.do_qsm and run_args.masking_input == 'magnitude':
        logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot use magnitude-based masking - unequal number of phase and magnitude files found!")
        run_args.masking_input = 'phase'
        run_args.masking_algorithm = 'threshold'
        run_args.inhomogeneity_correction = False
        run_args.add_bet = False
    if len(magnitude_files) == 1:
        if run_args.do_r2starmap:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot compute R2* - at least two echoes are needed!")
            run_args.do_r2starmap = False
        if run_args.do_t2starmap:
            logger.log(LogLevel.WARNING.value, f"{run_id}: Cannot compute T2* - at least two echoes are needed!")
            run_args.do_t2starmap = False

    if run_args.do_qsm or run_args.do_swi:
        if not all(all(nib.load(phase_files[i]).header['dim'] == nib.load(phase_files[0]).header['dim']) for i in range(len(phase_files))):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot do QSM or SWI - dimensions of phase files are unequal!")
            run_args.do_qsm = False
            run_args.do_swi = False
    if run_args.do_qsm and (run_args.masking_input == 'magnitude' or run_args.inhomogeneity_correction or run_args.add_bet):
        mag_dims = [nib.load(magnitude_files[i]).header['dim'][1:4].tolist() for i in range(len(magnitude_files))]
        phs_dims = [nib.load(magnitude_files[i]).header['dim'][1:4].tolist() for i in range(len(phase_files))]
        if not all(x == phs_dims[0] for x in mag_dims):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot use magnitude for masking - dimensions of magnitude files are not all equal to phase files! magnitude={mag_dims}; phase={phs_dims}.")
            run_args.masking_input = 'phase'
            run_args.inhomogeneity_correction = False
            run_args.add_bet = False
        if run_args.use_existing_masks:
            mask_dims = [nib.load(mask_files[i]).header['dim'][1:4].tolist() for i in range(len(mask_files))]
            phs_dims = [nib.load(magnitude_files[i]).header['dim'][1:4].tolist() for i in range(len(phase_files))]
            if not all(x == phs_dims[0] for x in mask_dims):
                logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot use existing masks - mask dimensions are not all equal to phase files! mask={mask_dims}; phase={phs_dims}.")
                run_args.use_existing_masks = False
    elif run_args.do_r2starmap or run_args.do_t2starmap:
        mag_dims = [nib.load(magnitude_files[i]).header['dim'][1:4].tolist() for i in range(len(magnitude_files))]
        if not all(x == mag_dims[0] for x in mag_dims):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot do T2*/R2* mapping - magnitude dimensions are not all equal! magnitude={mag_dims}.")
            run_args.do_r2starmap = False
            run_args.do_t2starmap = False
    if run_args.do_qsm or run_args.do_swi:
        if any(nib.load(phase_files[i]).header['dim'][0] > 3 for i in range(len(phase_files))):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot do QSM or SWI - >3D phase files detected! Each volume must be 3D, coil-combined, and represent a single echo for BIDS-compliance.")
            run_args.do_qsm = False
            run_args.do_swi = False
    if (run_args.do_qsm and (run_args.masking_input == 'magnitude' or run_args.inhomogeneity_correction or run_args.add_bet)) or run_args.do_r2starmap or run_args.do_t2starmap:
        if any(nib.load(magnitude_files[i]).header['dim'][0] > 3 for i in range(len(magnitude_files))):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot do magnitude-based masking - >3D magnitude files detected! Each volume must be 3D, coil-combined, and represent a single echo for BIDS-compliance.")
            run_args.masking_input = 'phase'
            run_args.inhomogeneity_correction = False
            run_args.add_bet = False
            run_args.do_r2starmap = False
            run_args.do_t2starmap = False
    if (run_args.do_qsm and run_args.use_existing_masks):
        if any(nib.load(mask_files[i]).header['dim'][0] > 3 for i in range(len(mask_files))):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Cannot use existing masks - >3D masks detected! Each mask must be 3D only for BIDS-compliance.")
            run_args.use_existing_masks = False
    if run_args.do_analysis and not (qsm_files or run_args.do_qsm):
        logger.log(LogLevel.WARNING.value, f"{run_id}: Skipping analysis - no QSM files found or --do_qsm not selected!")
        run_args.do_analysis = False
    if run_args.do_analysis and not (seg_files or run_args.do_segmentation):
        logger.log(LogLevel.WARNING.value, f"{run_id}: Skipping analysis - no segmentations found or --do_segmentation not selected!")
        run_args.do_analysis = False

    # ensure that all input files dimensions match and are 3d
    if phase_files and any([run_args.do_qsm, run_args.do_swi]):
        dims = nib.load(phase_files[0]).header.get_data_shape()
        if len(dims) != 3 or any(dim == 1 for dim in dims[:3]):
            logger.log(LogLevel.ERROR.value, f"{run_id}: Input dimensions must be 3D! Got {phase_files[0]}={dims}.")
        for i in range(1, len(phase_files)):
            dims_i = nib.load(phase_files[i]).header.get_data_shape()
            if dims != dims_i:
                logger.log(LogLevel.ERROR.value, f"{run_id}: Incompatible dimensions detected! {phase_files[0]}={dims}; {phase_files[i]}={dims}.")
        
        if any([run_args.do_t2starmap, run_args.do_r2starmap, run_args.do_segmentation, run_args.masking_input == 'magnitude', run_args.inhomogeneity_correction, run_args.add_bet]):
            if not run_args.do_qsm:
                dims = nib.load(magnitude_files[0]).header.get_data_shape()
            if len(dims) != 3 or any(dim == 1 for dim in dims[:3]):
                logger.log(LogLevel.ERROR.value, f"{run_id}: Input dimensions must be 3D! Got {magnitude_files[0]}={dims}.")
            for i in range(len(magnitude_files)):
                dims_i = nib.load(magnitude_files[i]).header.get_data_shape()
                if dims != dims_i:
                    logger.log(LogLevel.ERROR.value, f"{run_id}: Incompatible dimensions detected! {phase_files[0] if run_args.do_qsm else magnitude_files[0]}={dims}; {magnitude_files[i]}={dims}.")
        
        if run_args.do_qsm and run_args.use_existing_masks and mask_files:
            dims = nib.load(phase_files[0]).header.get_data_shape()
            for i in range(len(mask_files)):
                dims_i = nib.load(mask_files[i]).header.get_data_shape()
                if dims != dims_i:
                    logger.log(LogLevel.ERROR.value, f"{run_id}: Incompatible dimensions detected! {phase_files[0]}={dims}; {mask_files[i]}={dims}.")
    
    def calculate_memory_usage(nifti_file):
        nifti = nib.load(nifti_file)
        header = nifti.header
        dimensions = header.get_data_shape()
        num_voxels = np.prod(dimensions)
        bytepix = header.get_data_dtype().itemsize
        memory_usage_bytes = num_voxels * bytepix
        mem_gb = memory_usage_bytes / (1024 ** 3)
        return mem_gb, dimensions, bytepix
    
    if phase_files:
        mem_phase, dimensions_phase, bytepix_phase = calculate_memory_usage(phase_files[0])
        mem_phase_64 = np.prod(dimensions_phase) * 8 / (1024 ** 3)
        logger.log(LogLevel.DEBUG.value, f"GRE phase files are {dimensions_phase} * {len(phase_files)} echoes * {bytepix_phase} bytes/voxel == {round(mem_phase * len(phase_files), 3)} GB ({round(mem_phase_64 * len(phase_files), 3)} GB at 64-bit)")
    if magnitude_files:
        mem_mag, dimensions_mag, bytepix_mag = calculate_memory_usage(magnitude_files[0])
        mem_mag_64 = np.prod(dimensions_mag) * 8 / (1024 ** 3)
        logger.log(LogLevel.DEBUG.value, f"GRE magnitude files are {dimensions_mag} * {len(magnitude_files)} echoes * {bytepix_mag} bytes/voxel == {round(mem_mag * len(magnitude_files), 3)} GB ({round(mem_mag_64 * len(magnitude_files), 3)} GB at 64-bit)")
    if t1w_files:
        mem_t1w, dimensions_t1w, bytepix_t1w = calculate_memory_usage(t1w_files[0])
        mem_t1w_64 = np.prod(dimensions_t1w) * 8 / (1024 ** 3)
        logger.log(LogLevel.DEBUG.value, f"T1w files are {dimensions_t1w} * {bytepix_t1w} bytes/voxel == {round(mem_t1w, 3)} GB ({round(mem_t1w_64, 3)} GB at 64-bit)")
    if mask_files and run_args.use_existing_masks:
        mem_mask, dimensions_mask, bytepix_mask = calculate_memory_usage(mask_files[0])
        mem_mask_64 = np.prod(dimensions_mask) * 8 / (1024 ** 3)
        logger.log(LogLevel.DEBUG.value, f"Mask files are {dimensions_mask} * {bytepix_mask} bytes/voxel == {round(mem_mask, 3)} GB ({round(mem_mask_64 * len(mask_files), 3)} GB at 64-bit)")

    if not any([run_args.do_qsm, run_args.do_swi, run_args.do_t2starmap, run_args.do_r2starmap, run_args.do_segmentation, run_args.do_analysis]):
        return
    
    # create nipype workflow for this run
    wf = Workflow(f"qsmxt" + (f"_acq-{acq}" if acq else "") + (f"_run-{run}" if run else ""), base_dir=os.path.join(run_args.output_dir, "workflow", os.path.join(subject, session) if session else subject, acq or "", run or ""))

    # inputs and outputs
    n_inputs = create_node(
        IdentityInterface(
            fields=['phase', 'magnitude', 'params_files', 'mask']
        ),
        name='nipype_getfiles'
    )
    n_inputs.inputs.phase = phase_files[0] if len(phase_files) == 1 else phase_files
    n_inputs.inputs.magnitude = magnitude_files[0] if len(magnitude_files) == 1 else magnitude_files
    n_inputs.inputs.params_files = params_files[0] if len(params_files) == 1 else params_files
    if not run_args.combine_phase and len(phase_files) > 1 and len(mask_files) == 1:
        mask_files = [mask_files[0] for x in phase_files]
        n_inputs.inputs.mask = mask_files
    else:
        n_inputs.inputs.mask = mask_files[0] if len(mask_files) == 1 else mask_files

    n_outputs = create_node(
        IdentityInterface(
            fields=['qsm', 'qsm_singlepass', 'swi', 'swi_mip', 't2s', 'r2s', 't1w_segmentation', 'qsm_segmentation', 'transform', 'analysis_csv', 'qsm_dicoms', 'swi_dicoms', 'swi_mip_dicoms']
        ),
        name='qsmxt_outputs'
    )
    n_copyfile = Node(copyfile.DynamicCopyFiles(infields=[
        'qsm', 'qsm_singlepass', 'swi', 'swi_mip', 't2s', 'r2s', 't1w_segmentation', 'qsm_segmentation', 'transform', 'analysis_csv', 'qsm_dicoms', 'swi_dicoms', 'swi_mip_dicoms'
    ]), name="copyfile")
    
    basedir = os.path.join(run_args.output_dir, subject, session if session else '')
    basename = f"{subject}"
    if session: basename += f"_{session}"
    if acq: basename += f"_acq-{acq}"
    if run: basename += f"_run-{run}"
    
    n_copyfile.inputs.output_map = {
        'qsm': os.path.join(basedir, 'anat', f"{basename}_Chimap"),
        'qsm_singlepass': os.path.join(basedir, 'anat', f"{basename}_desc-singlepass_Chimap"),
        'swi': os.path.join(basedir, 'anat', f"{basename}_swi"),
        'swi_mip': os.path.join(basedir, 'anat', f"{basename}_minIP"),
        't2s': os.path.join(basedir, 'anat', f"{basename}_T2starmap"),
        'r2s': os.path.join(basedir, 'anat', f"{basename}_R2starmap"),
        't1w_segmentation': os.path.join(basedir, 'anat', f"{basename}_space-orig_dseg"),
        'qsm_segmentation': os.path.join(basedir, 'anat', f"{basename}_space-qsm_dseg"),
        'transform': os.path.join(basedir, 'extra_data', f"{basename}_desc-t1w-to-qsm_transform"),
        'analysis_csv': os.path.join(basedir, 'extra_data', f"{basename}_qsm-analysis"),
        'qsm_dicoms': os.path.join(basedir, 'extra_data', f"{basename}_desc-dicoms_Chimap"),
        'swi_dicoms': os.path.join(basedir, 'extra_data', f"{basename}_desc-dicoms_swi"),
        'swi_mip_dicoms': os.path.join(basedir, 'extra_data', f"{basename}_desc-dicoms_minIP")
    }
    
    wf.connect([
        (n_outputs, n_copyfile, [
            ('qsm', 'qsm'),
            ('qsm_singlepass', 'qsm_singlepass'),
            ('swi', 'swi'),
            ('swi_mip', 'swi_mip'),
            ('t2s', 't2s'),
            ('r2s', 'r2s'),
            ('t1w_segmentation', 't1w_segmentation'),
            ('qsm_segmentation', 'qsm_segmentation'),
            ('transform', 'transform'),
            ('analysis_csv', 'analysis_csv'),
            ('qsm_dicoms', 'qsm_dicoms'),
            ('swi_dicoms', 'swi_dicoms'),
            ('swi_mip_dicoms', 'swi_mip_dicoms')
        ])
    ])
    # read echotime and field strengths from json files
    def read_json_me(params_file):
        import json
        with open(params_file, 'rt') as json_file:
            data = json.load(json_file)
        te = data['EchoTime']
        json_file.close()
        return te
    def read_json_se(params_files):
        import json
        with open(params_files[0] if isinstance(params_files, list) else params_files, 'rt') as json_file:
            data = json.load(json_file)
        B0 = data['MagneticFieldStrength']
        json_file.close()
        return B0
    mn_json_params = create_node(
        interface=Function(
            input_names=['params_file'],
            output_names=['TE'],
            function=read_json_me
        ),
        iterfield=['params_file'],
        name='func_read-json-me',
        is_map=len(params_files) > 1
    )
    wf.connect([
        (n_inputs, mn_json_params, [('params_files', 'params_file')])
    ])
    n_json_params = create_node(
        interface=Function(
            input_names=['params_files'],
            output_names=['B0'],
            function=read_json_se
        ),
        iterfield=['params_file'],
        name='func_read-json-se'
    )
    wf.connect([
        (n_inputs, n_json_params, [('params_files', 'params_files')])
    ])

    # read voxel size 'vsz' from nifti file
    def read_nii(nii_file):
        import nibabel as nib
        import numpy as np
        if isinstance(nii_file, list): nii_file = nii_file[0]
        nii = nib.load(nii_file)
        return np.array(nii.header.get_zooms()).tolist()
    n_nii_params = create_node(
        interface=Function(
            input_names=['nii_file'],
            output_names=['vsz'],
            function=read_nii
        ),
        name='nibabel_read-nii'
    )
    wf.connect([
        (n_inputs, n_nii_params, [('phase' if phase_files else 'magnitude', 'nii_file')])
    ])

    # reorient to canonical
    def as_closest_canonical(phase=None, magnitude=None, mask=None):
        import os
        import nibabel as nib
        from qsmxt.scripts.qsmxt_functions import extend_fname

        assert(phase or magnitude or mask)

        def as_closest_canonical_i(in_file):
            if nib.aff2axcodes(nib.load(in_file).affine) == ('R', 'A', 'S'):
                return in_file
            else:
                out_file = extend_fname(in_file, "_canonical", out_dir=os.getcwd())
                nib.save(nib.as_closest_canonical(nib.load(in_file)), out_file)
                return out_file
        
        out_phase = None
        out_mag = None
        out_mask = None

        if phase:
            if isinstance(phase, list):
                out_phase = [as_closest_canonical_i(phase_i) for phase_i in phase]
            else:
                out_phase = as_closest_canonical_i(phase)
        if magnitude:
            if isinstance(magnitude, list):
                out_mag = [as_closest_canonical_i(magnitude_i) for magnitude_i in magnitude]
            else:
                out_mag = as_closest_canonical_i(magnitude)
        if mask:
            if isinstance(mask, list):
                out_mask = [as_closest_canonical_i(mask_i) for mask_i in mask]
            else:
                out_mask = as_closest_canonical_i(mask)
        
        return out_phase, out_mag, out_mask
    mn_inputs_canonical = create_node(
        interface=Function(
            input_names=[] + (['phase'] if (phase_files and (run_args.do_swi or run_args.do_qsm)) else []) + (['magnitude'] if magnitude_files else []) + (['mask'] if mask_files and run_args.use_existing_masks else []),
            output_names=['phase', 'magnitude', 'mask'],
            function=as_closest_canonical
        ),
        #iterfield=['phase'] + (['magnitude'] if magnitude_files else []) + (['mask'] if mask_files else []),
        name='nibabel_as-canonical',
        #is_map=len(phase_files) > 1
    )
    if phase_files and (run_args.do_swi or run_args.do_qsm):
        wf.connect([
            (n_inputs, mn_inputs_canonical, [('phase', 'phase')]),
        ])
    if magnitude_files:
        wf.connect([
            (n_inputs, mn_inputs_canonical, [('magnitude', 'magnitude')]),
        ])
    if mask_files and run_args.use_existing_masks:
        wf.connect([
            (n_inputs, mn_inputs_canonical, [('mask', 'mask')]),
        ])

    # scale phase
    if phase_files and (run_args.do_swi or run_args.do_qsm):
        mn_phase_scaled = create_node(
            interface=processphase.ScalePhaseInterface(),
            iterfield=['phase'],
            name='nibabel_numpy_scale-phase',
            is_map=len(phase_files) > 1
        )
        wf.connect([
            (mn_inputs_canonical, mn_phase_scaled, [('phase', 'phase')])
        ])

    # r2* and t2* mappping
    if run_args.do_t2starmap or run_args.do_r2starmap:
        n_t2s_r2s_mem = mem_mag_64 * (len(magnitude_files) + 2)
        n_t2s_r2s = create_node(
            interface=t2s_r2s.T2sR2sInterface(),
            mem_gb=n_t2s_r2s_mem,
            name='mrt_t2s-r2s'
        )
        n_t2s_r2s.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="t2s-r2s",
            time="01:00:00",
            mem_gb=n_t2s_r2s_mem
        )
        wf.connect([
            (mn_inputs_canonical, n_t2s_r2s, [('magnitude', 'magnitude')]),
            (mn_json_params, n_t2s_r2s, [('TE', 'TE')])
        ])
        if run_args.do_t2starmap: wf.connect([(n_t2s_r2s, n_outputs, [('t2starmap', 't2s')])])
        if run_args.do_r2starmap: wf.connect([(n_t2s_r2s, n_outputs, [('r2starmap', 'r2s')])])
    
    # swi
    if run_args.do_swi:
        n_swi_mem = mem_mag_64 * (len(magnitude_files) + 2) + mem_phase_64 * len(phase_files)
        n_swi_threads = min(6, run_args.n_procs) if run_args.multiproc else 6
        n_swi = create_node(
            interface=swi.ClearSwiInterface(
                num_threads=n_swi_threads
            ),
            name='mrt_clearswi',
            mem_gb=n_swi_mem,
            n_procs=n_swi_threads
        )
        n_swi.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="SWI",
            mem_gb=n_swi_mem,
            num_cpus=n_swi_threads
        )
        wf.connect([
            (mn_phase_scaled, n_swi, [('phase', 'phase')]),
            (mn_inputs_canonical, n_swi, [('magnitude', 'magnitude')]),
            (mn_json_params, n_swi, [('TE', 'TEs' if len(phase_files) > 1 else 'TE')]),
            (n_swi, n_outputs, [('swi', 'swi')]),
            (n_swi, n_outputs, [('swi_mip', 'swi_mip')])
        ])

    # segmentation
    if run_args.do_segmentation:
        n_registration_threads = min(6, run_args.n_procs) if run_args.multiproc else 6
        n_registration_mem = 8
        n_registration = create_node(
            interface=RegistrationSynQuick(
                num_threads=n_registration_threads,
                fixed_image=magnitude_files[0],
                moving_image=t1w_files[0],
                output_prefix=f"{subject}_{session}" + (f"_acq-{acq}" if acq else "") + (f"_run-{run}" if run else "") + "_"
            ),
            name='ants_register-t1-to-qsm',
            n_procs=n_registration_threads,
            mem_gb=n_registration_mem
        )
        n_registration.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="ANTS",
            mem_gb=n_registration_mem,
            num_cpus=n_registration_threads
        )

        # segment t1
        n_fastsurfer_threads = min(8, run_args.n_procs) if run_args.multiproc else 8
        n_fastsurfer_mem = 12
        n_fastsurfer = create_node(
            interface=fastsurfer.FastSurferInterface(
                in_file=t1w_files[0],
                num_threads=n_fastsurfer_threads
            ),
            name='fastsurfer_segment-t1',
            n_procs=n_fastsurfer_threads,
            mem_gb=n_fastsurfer_mem
        )
        n_fastsurfer.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="FASTSURFER",
            mem_gb=n_fastsurfer_mem,
            num_cpus=n_fastsurfer_threads
        )

        # convert segmentation to nii
        n_fastsurfer_aseg_nii = create_node(
            interface=mgz2nii.Mgz2NiiInterface(),
            name='numpy_numpy_nibabel_mgz2nii'
        )
        wf.connect([
            (n_fastsurfer, n_fastsurfer_aseg_nii, [('out_file', 'in_file')])
        ])

        # resample segmentation in T1w space
        n_fastsurfer_aseg_nii_resampled = create_node(
            interface=resample_like.ResampleLikeInterface(
                ref_file=t1w_files[0],
                interpolation='nearest'
            ),
            name='nibabel_numpy_nilearn_t1w-seg-resampled'
        )
        wf.connect([
            (n_fastsurfer_aseg_nii, n_fastsurfer_aseg_nii_resampled, [('out_file', 'in_file')])
        ])

        # get first canonical magnitude
        n_getfirst_canonical_magnitude = create_node(
            interface=Function(
                input_names=['magnitude'],
                output_names=['magnitude'],
                function=lambda magnitude: magnitude[0] if isinstance(magnitude, list) else magnitude
            ),
            name='func_getfirst-canonical-magnitude'
        )
        wf.connect([
            (mn_inputs_canonical, n_getfirst_canonical_magnitude, [('magnitude', 'magnitude')])
        ])

        # apply transforms to segmentation
        n_transform_segmentation = create_node(
            interface=ApplyTransforms(
                dimension=3,
                interpolation="NearestNeighbor",
                output_image=f"{run_id.replace('.', '_')}_segmentation_trans.nii"
            ),
            name='ants_transform-segmentation-to-qsm'
        )
        wf.connect([
            (n_fastsurfer_aseg_nii, n_transform_segmentation, [('out_file', 'input_image')]),
            (n_getfirst_canonical_magnitude, n_transform_segmentation, [('magnitude', 'reference_image')]),
            (n_registration, n_transform_segmentation, [('out_matrix', 'transforms')])
        ])

        wf.connect([
            (n_fastsurfer_aseg_nii_resampled, n_outputs, [('out_file', 't1w_segmentation')]),
            (n_transform_segmentation, n_outputs, [('output_image', 'qsm_segmentation')]),
            (n_registration, n_outputs, [('out_matrix', 'transform')])
        ])

    if run_args.do_qsm:
        
        # resample to axial
        n_inputs_resampled = create_node(
            interface=IdentityInterface(
                fields=['phase', 'magnitude', 'mask']
            ),
            name='nipype_inputs-resampled'
        )
        if magnitude_files:
            mn_resample_mem = (np.prod(dimensions_phase) * bytepix_phase) / (1024 ** 3) * 2 * 16
            mn_resample_inputs = create_node(
                interface=sampling.AxialSamplingInterface(
                    obliquity_threshold=999 if run_args.obliquity_threshold == -1 else run_args.obliquity_threshold
                ),
                iterfield=['magnitude', 'phase'],
                mem_gb=mn_resample_mem,
                name='nibabel_numpy_nilearn_axial-resampling',
                is_map=len(phase_files) > 1
            )
            mn_resample_inputs.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=run_args.slurm[0],
                pbs_account=run_args.pbs,
                slurm_partition=run_args.slurm[1],
                name="axial_resampling",
                time="00:10:00",
                mem_gb=mn_resample_mem
            )
            wf.connect([
                (mn_inputs_canonical, mn_resample_inputs, [('magnitude', 'magnitude')]),
                (mn_phase_scaled, mn_resample_inputs, [('phase', 'phase')]),
                (mn_resample_inputs, n_inputs_resampled, [('magnitude', 'magnitude')]),
                (mn_resample_inputs, n_inputs_resampled, [('phase', 'phase')])
            ])
            if mask_files and run_args.use_existing_masks:
                mn_resample_mask_mem = (np.prod(dimensions_phase) * bytepix_phase) / (1024 ** 3) * 16
                mn_resample_mask = create_node(
                    interface=sampling.AxialSamplingInterface(
                        obliquity_threshold=999 if run_args.obliquity_threshold == -1 else run_args.obliquity_threshold
                    ),
                    iterfield=['mask'],
                    mem_gb=mn_resample_mask_mem,
                    name='nibabel_numpy_nilearn_axial-resampling-mask',
                    is_map=isinstance(n_inputs.inputs.mask, list) and len(n_inputs.inputs.mask) > 1
                )
                mn_resample_mask.plugin_args = gen_plugin_args(
                    plugin_args={ 'overwrite': True },
                    slurm_account=run_args.slurm[0],
                    pbs_account=run_args.pbs,
                    slurm_partition=run_args.slurm[1],
                    name="axial_resampling",
                    time="00:10:00",
                    mem_gb=mn_resample_mask_mem,
                )
                wf.connect([
                    (mn_inputs_canonical, mn_resample_mask, [('mask', 'mask')]),
                    (mn_resample_mask, n_inputs_resampled, [('mask', 'mask')])
                ])
        else:
            wf.connect([
                (mn_phase_scaled, n_inputs_resampled, [('phase', 'phase')])
            ])
            if mask_files:
                wf.connect([
                    (mn_inputs_canonical, n_inputs_resampled, [('mask', 'mask')])
                ])

        # combine phase data if necessary
        n_inputs_combine = create_node(
            interface=IdentityInterface(
                fields=['phase_unwrapped', 'frequency']
            ),
            name='phase-combined'
        )
        if run_args.combine_phase:
            n_romeo_mem = (54.0860 * (np.prod(dimensions_phase) * 8 / (1024**3)) + 3.9819) # DONE
            n_romeo_combine = create_node(
                interface=romeo.RomeoB0Interface(),
                name='mrt_romeo_combine-phase',
                mem_gb=n_romeo_mem
            )
            n_romeo_combine.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=run_args.slurm[0],
                pbs_account=run_args.pbs,
                slurm_partition=run_args.slurm[1],
                name="romeo_combine",
                time="00:10:00",
                mem_gb=n_romeo_mem,
            )
            wf.connect([
                (mn_json_params, n_romeo_combine, [('TE', 'TEs')]),
                (n_inputs_resampled, n_romeo_combine, [('phase', 'phase')]),
                (n_inputs_resampled, n_romeo_combine, [('magnitude', 'magnitude')]),
                (n_romeo_combine, n_inputs_combine, [('frequency', 'frequency')]),
                (n_romeo_combine, n_inputs_combine, [('phase_unwrapped', 'phase_unwrapped')]),
            ])

        # === MASKING ===
        wf_masking = masking_workflow(
            run_args=run_args,
            mask_available=len(mask_files) > 0 and run_args.use_existing_masks,
            magnitude_available=len(magnitude_files) > 0,
            qualitymap_available=False,
            fill_masks=True,
            add_bet=run_args.add_bet and run_args.filling_algorithm != 'bet',
            use_maps=len(phase_files) > 1 and not run_args.combine_phase,
            name="mask",
            dimensions_phase=dimensions_phase,
            bytepix_phase=bytepix_phase,
            num_echoes=len(magnitude_files),
            index=0
        )
        wf.connect([
            (n_inputs_resampled, wf_masking, [('phase', 'masking_inputs.phase')]),
            (n_inputs_resampled, wf_masking, [('magnitude', 'masking_inputs.magnitude')]),
            (n_inputs_resampled, wf_masking, [('mask', 'masking_inputs.mask')]),
            (mn_json_params, wf_masking, [('TE', 'masking_inputs.TE')])
        ])

        # === QSM ===
        wf_qsm = qsm_workflow(run_args, "qsm", len(magnitude_files) > 0, len(phase_files) > 1 and not run_args.combine_phase, dimensions_phase, bytepix_phase, qsm_erosions=run_args.tgv_erosions)

        wf.connect([
            (n_inputs_resampled, wf_qsm, [('phase', 'qsm_inputs.phase')]),
            (n_inputs_combine, wf_qsm, [('phase_unwrapped', 'qsm_inputs.phase_unwrapped')]),
            (n_inputs_combine, wf_qsm, [('frequency', 'qsm_inputs.frequency')]),
            (n_inputs_resampled, wf_qsm, [('magnitude', 'qsm_inputs.magnitude')]),
            (wf_masking, wf_qsm, [('masking_outputs.mask', 'qsm_inputs.mask')]),
            (mn_json_params, wf_qsm, [('TE', 'qsm_inputs.TE')]),
            (n_json_params, wf_qsm, [('B0', 'qsm_inputs.B0')]),
            (n_nii_params, wf_qsm, [('vsz', 'qsm_inputs.vsz')]),
        ])
        wf_qsm.get_node('qsm_inputs').inputs.b0_direction = [0, 0, 1]
        
        n_qsm_average = create_node(
            interface=nonzeroaverage.NonzeroAverageInterface(),
            name="nibabel_numpy_qsm-average"
        )
        wf.connect([
            (wf_qsm, n_qsm_average, [('qsm_outputs.qsm', 'in_files')]),
            (wf_masking, n_qsm_average, [('masking_outputs.mask', 'in_masks')])
        ])

        n_resample_qsm = create_node(
            interface=resample_like.ResampleLikeInterface(),
            name='nibabel_numpy_nilearn_qsm-resampled'
        )
        wf.connect([
            (n_qsm_average, n_resample_qsm, [('out_file', 'in_file')]),
            (mn_inputs_canonical, n_resample_qsm, [('phase', 'ref_file')])
        ])

        if run_args.qsm_reference:
            n_qsm_referenced = create_node(
                interface=qsm_referencing.ReferenceQSMInterface(
                    in_seg_values=run_args.qsm_reference if isinstance(run_args.qsm_reference, list) and run_args.do_segmentation else None
                ),
                name='nibabel_numpy_qsm-referenced'
            )
            wf.connect([
                (n_resample_qsm, n_qsm_referenced, [('out_file', 'in_qsm')]),
                (n_qsm_referenced, n_outputs, [('out_file', 'qsm' if not run_args.two_pass else 'qsm_singlepass')])
            ])
            if isinstance(run_args.qsm_reference, list) and run_args.do_segmentation:
                wf.connect([
                    (n_transform_segmentation, n_qsm_referenced, [('output_image', 'in_seg')])
                ])
        else:
            wf.connect([
                (n_resample_qsm, n_outputs, [('out_file', 'qsm' if not run_args.two_pass else 'qsm_singlepass')])
            ])

        # two-pass algorithm
        if run_args.two_pass:
            wf_masking_intermediate = masking_workflow(
                run_args=run_args,
                mask_available=False,
                magnitude_available=len(magnitude_files) > 0,
                qualitymap_available=True,
                fill_masks=False,
                add_bet=False,
                use_maps=len(phase_files) > 1 and not run_args.combine_phase,
                name="mask-intermediate",
                dimensions_phase=dimensions_phase,
                bytepix_phase=bytepix_phase,
                num_echoes=len(magnitude_files),
                index=1
            )
            wf.connect([
                (n_inputs_resampled, wf_masking_intermediate, [('phase', 'masking_inputs.phase')]),
                (n_inputs_resampled, wf_masking_intermediate, [('magnitude', 'masking_inputs.magnitude')]),
                (n_inputs_resampled, wf_masking_intermediate, [('mask', 'masking_inputs.mask')]),
                (mn_json_params, wf_masking_intermediate, [('TE', 'masking_inputs.TE')]),
                (wf_masking, wf_masking_intermediate, [('masking_outputs.quality_map', 'masking_inputs.quality_map')])
            ])

            wf_qsm_intermediate = qsm_workflow(run_args, "qsm-intermediate", len(magnitude_files) > 0, len(phase_files) > 1 and not run_args.combine_phase, dimensions_phase, bytepix_phase, qsm_erosions=0)
            wf.connect([
                (n_inputs_resampled, wf_qsm_intermediate, [('phase', 'qsm_inputs.phase')]),
                (n_inputs_resampled, wf_qsm_intermediate, [('magnitude', 'qsm_inputs.magnitude')]),
                (n_inputs_combine, wf_qsm_intermediate, [('phase_unwrapped', 'qsm_inputs.phase_unwrapped')]),
                (n_inputs_combine, wf_qsm_intermediate, [('frequency', 'qsm_inputs.frequency')]),
                (mn_json_params, wf_qsm_intermediate, [('TE', 'qsm_inputs.TE')]),
                (n_json_params, wf_qsm_intermediate, [('B0', 'qsm_inputs.B0')]),
                (n_nii_params, wf_qsm_intermediate, [('vsz', 'qsm_inputs.vsz')]),
                (wf_masking_intermediate, wf_qsm_intermediate, [('masking_outputs.mask', 'qsm_inputs.mask')])
            ])
            wf_qsm_intermediate.get_node('qsm_inputs').inputs.b0_direction = [0, 0, 1]
                    
            # two-pass combination
            mn_qsm_twopass = create_node(
                interface=twopass.TwopassNiftiInterface(),
                name='numpy_nibabel_twopass',
                iterfield=['in_file', 'in_filled', 'mask'],
                is_map=len(phase_files) > 1 and not run_args.combine_phase
            )
            wf.connect([
                (wf_qsm_intermediate, mn_qsm_twopass, [('qsm_outputs.qsm', 'in_file')]),
                (wf_masking_intermediate, mn_qsm_twopass, [('masking_outputs.mask', 'mask')]),
                (wf_qsm, mn_qsm_twopass, [('qsm_outputs.qsm', 'in_filled')])
            ])

            # averaging
            n_qsm_twopass_average = create_node(
                interface=nonzeroaverage.NonzeroAverageInterface(),
                name="nibabel_numpy_twopass-average"
            )
            wf.connect([
                (mn_qsm_twopass, n_qsm_twopass_average, [('out_file', 'in_files')]),
                #(wf_masking, n_qsm_twopass_average, [('masking_outputs.mask', 'in_masks')])
            ])

            n_resample_qsm = create_node(
                interface=resample_like.ResampleLikeInterface(),
                name='nibabel_numpy_nilearn_twopass-qsm-resampled'
            )
            wf.connect([
                (n_qsm_twopass_average, n_resample_qsm, [('out_file', 'in_file')]),
                (mn_inputs_canonical, n_resample_qsm, [('phase', 'ref_file')])
            ])

            if run_args.qsm_reference:
                n_qsm_twopass_referenced = create_node(
                    interface=qsm_referencing.ReferenceQSMInterface(
                        in_seg_values=run_args.qsm_reference if isinstance(run_args.qsm_reference, list) and run_args.do_segmentation else None
                    ),
                    name='nibabel_numpy_qsm-twopass-referenced'
                )
                wf.connect([
                    (n_resample_qsm, n_qsm_twopass_referenced, [('out_file', 'in_qsm')]),
                    (n_qsm_twopass_referenced, n_outputs, [('out_file', 'qsm')])
                ])
                if isinstance(run_args.qsm_reference, list) and run_args.do_segmentation:
                    wf.connect([
                        (n_transform_segmentation, n_qsm_twopass_referenced, [('output_image', 'in_seg')])
                    ])
            else:
                wf.connect([
                    (n_resample_qsm, n_outputs, [('out_file', 'qsm')])
                ])

    if run_args.do_analysis:
        def combine_lists(list1=None, list2=None):
            if list1 is None:
                return list2
            if list2 is None:
                return list1
            if not isinstance(list1, list): list1 = [list1]
            if not isinstance(list2, list): list2 = [list2]
            return list1 + list2
        n_combine_qsm_files = create_node(
            interface=Function(input_names=['list1', 'list2'], output_names=['qsm_files'], function=combine_lists),
            name='combine_lists1'
        )
        n_combine_seg_files = create_node(
            interface=Function(input_names=['list1', 'list2'], output_names=['seg_files'], function=combine_lists),
            name='combine_lists2'
        )
        n_combine_qsm_files.inputs.list1 = qsm_files
        n_combine_seg_files.inputs.list1 = seg_files
        if run_args.do_qsm:
            wf.connect([
                (n_resample_qsm, n_combine_qsm_files, [('out_file', 'list2')]),
            ])
        if run_args.do_segmentation:
            wf.connect([
                (n_transform_segmentation, n_combine_seg_files, [('output_image', 'list2')]),
            ])

        def create_combinations(list1, list2):
            import itertools
            combinations = list(itertools.product(list1, list2))
            in_files, in_segmentations = zip(*combinations)
            return list(in_files), list(in_segmentations)
        n_create_permutations = create_node(
            interface=Function(input_names=['list1', 'list2'], output_names=['qsm_files', 'seg_files'], function=create_combinations),
            name='create_permutations'
        )
        wf.connect(n_combine_qsm_files, 'qsm_files', n_create_permutations, 'list1')
        wf.connect(n_combine_seg_files, 'seg_files', n_create_permutations, 'list2')


        n_analyse_qsm = create_node(
            interface=analyse.AnalyseInterface(
                in_labels=run_args.labels_file,
                in_pipeline_name=os.path.split(run_args.output_dir)[1]
            ),
            name='nibabel_numpy_analyse-qsm',
            iterfield=['in_file', 'in_segmentation'],
            is_map=True
        )
        wf.connect(n_create_permutations, 'qsm_files', n_analyse_qsm, 'in_file')
        wf.connect(n_create_permutations, 'seg_files', n_analyse_qsm, 'in_segmentation')
        wf.connect(n_analyse_qsm, 'out_csv', n_outputs, 'analysis_csv')


    # insert DICOM conversion step
    if run_args.export_dicoms:
        for target_attribute in ['qsm', 'swi', 'swi_mip']:
            node, node_attribute = get_preceding_node_and_attribute(wf, target_node_name='qsmxt_outputs', target_attribute=target_attribute)
            if node:
                logger.log(LogLevel.DEBUG.value, f"Found node {node._name}")
                n_nii2dcm = create_node(
                    interface=nii2dcm.Nii2DcmInterface(centered=True if 'qsm' in target_attribute else False),
                    name=f'n_nii2dcm_{target_attribute}'
                )
                wf.connect([
                    (node, n_nii2dcm, [(node_attribute, 'in_file')]),
                    (n_nii2dcm, n_outputs, [('out_dir', f"{target_attribute}_dicoms")])
                ])
    
    return wf

def qsm_workflow(run_args, name, magnitude_available, use_maps, dimensions_phase, bytepix_phase, qsm_erosions=0):
    wf = Workflow(name=f"{name}_workflow")

    slurm_account = run_args.slurm[0] if run_args.slurm and len(run_args.slurm) else None
    slurm_partition = run_args.slurm[1] if run_args.slurm and len(run_args.slurm) > 1 else None

    n_inputs = create_node(
        interface=IdentityInterface(
            fields=['phase', 'phase_unwrapped', 'frequency', 'magnitude', 'mask', 'TE', 'B0', 'b0_direction', 'vsz']
        ),
        name='qsm_inputs'
    )

    n_outputs = create_node(
        interface=IdentityInterface(
            fields=['qsm', 'qsm_dicoms']
        ),
        name='qsm_outputs'
    )

    # === PHASE UNWRAPPING ===
    if run_args.unwrapping_algorithm:
        n_unwrapping = create_node(
            interface=IdentityInterface(
                fields=['phase_unwrapped']
            ),
            name='phase-unwrapping'
        )
        if run_args.unwrapping_algorithm == 'laplacian':
            laplacian_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
            laplacian_mem = 16.32256 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 1.12836 # DONE
            mn_laplacian = create_node(
                is_map=use_maps,
                interface=laplacian.LaplacianInterface(),
                iterfield=['phase'],
                name='mrt_laplacian-unwrapping',
                mem_gb=laplacian_mem,
                n_procs=laplacian_threads
            )
            wf.connect([
                (n_inputs, mn_laplacian, [('phase', 'phase')]),
                (mn_laplacian, n_unwrapping, [('phase_unwrapped', 'phase_unwrapped')])
            ])
            mn_laplacian.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=run_args.slurm[0],
                pbs_account=run_args.pbs,
                slurm_partition=run_args.slurm[1],
                name="Laplacian",
                mem_gb=laplacian_mem,
                num_cpus=laplacian_threads
            )
        if run_args.unwrapping_algorithm == 'romeo':
            if run_args.combine_phase:
                wf.connect([
                    (n_inputs, n_unwrapping, [('phase_unwrapped', 'phase_unwrapped')]),
                ])
            else:
                romeo_mem = 9.81512 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 1.75 # DONE
                mn_romeo = create_node(
                    interface=romeo.RomeoB0Interface(),
                    name='mrt_romeo',
                    mem_gb=romeo_mem
                )
                mn_romeo.plugin_args = gen_plugin_args(
                    plugin_args={ 'overwrite': True },
                    slurm_account=run_args.slurm[0],
                    pbs_account=run_args.pbs,
                    slurm_partition=run_args.slurm[1],
                    name="Romeo",
                    mem_gb=romeo_mem,
                )
                wf.connect([
                    (n_inputs, mn_romeo, [('phase', 'phase')]),
                    (n_inputs, mn_romeo, [('TE', 'TEs' if use_maps else 'TE')]),
                    (mn_romeo, n_unwrapping, [('phase_unwrapped', 'phase_unwrapped')])
                ])
                if magnitude_available:
                    wf.connect([
                        (n_inputs, mn_romeo, [('magnitude', 'magnitude')]),
                    ])


    # === PHASE TO FREQUENCY ===
    n_phase_normalized = create_node(
        interface=IdentityInterface(
            fields=['phase_normalized']
        ),
        name='phase_normalized'
    )
    if run_args.qsm_algorithm in ['rts', 'tv', 'nextqsm'] and not run_args.combine_phase:
        normalize_phase_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
        mn_normalize_phase = create_node(
            interface=processphase.PhaseToNormalizedInterface(
                scale_factor=1e6/(2*np.pi)
            ),
            name='nibabel-numpy_normalize-phase',
            iterfield=['phase', 'TE'],
            n_procs=normalize_phase_threads,
            is_map=use_maps
        )
        wf.connect([
            (n_unwrapping, mn_normalize_phase, [('phase_unwrapped', 'phase')]),
            (n_inputs, mn_normalize_phase, [('TE', 'TE')]),
            (n_inputs, mn_normalize_phase, [('B0', 'B0')]),
            (mn_normalize_phase, n_phase_normalized, [('phase_normalized', 'phase_normalized')])
        ])
        mn_normalize_phase.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="PhaToNormalized",
            num_cpus=normalize_phase_threads
        )
    if run_args.qsm_algorithm in ['rts', 'tv', 'nextqsm'] and run_args.combine_phase:
        normalize_freq_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
        mn_normalize_freq = create_node(
            interface=processphase.FreqToNormalizedInterface(
                scale_factor=1e6/(2*np.pi)
            ),
            name='nibabel-numpy_normalize-freq',
            iterfield=['frequency'],
            n_procs=normalize_freq_threads,
            is_map=use_maps
        )
        wf.connect([
            (n_inputs, mn_normalize_freq, [('frequency', 'frequency')]),
            (n_inputs, mn_normalize_freq, [('B0', 'B0')]),
            (mn_normalize_freq, n_phase_normalized, [('phase_normalized', 'phase_normalized')])
        ])
        mn_normalize_freq.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="PhaToNormalized",
            num_cpus=normalize_freq_threads
        )
    if run_args.qsm_algorithm in ['tgv', 'tgvjl'] and run_args.combine_phase:
        freq_to_phase_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
        mn_freq_to_phase = create_node(
            interface=processphase.FreqToPhaseInterface(TE=0.005, wraps=True),
            name='nibabel-numpy_freq-to-phase',
            iterfield=['frequency'],
            n_procs=freq_to_phase_threads,
            is_map=use_maps
        )
        wf.connect([
            (n_inputs, mn_freq_to_phase, [('frequency', 'frequency')]),
            (mn_freq_to_phase, n_phase_normalized, [('phase', 'phase_normalized')])
        ])
        mn_freq_to_phase.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="FreqToPhase",
            num_cpus=freq_to_phase_threads
        )
        
    # === BACKGROUND FIELD REMOVAL ===
    if run_args.qsm_algorithm in ['rts', 'tv']:
        mn_bf = create_node(
            interface=IdentityInterface(
                fields=['tissue_frequency', 'mask']
            ),
            name='bf-removal'
        )
        if run_args.bf_algorithm == 'vsharp':
            vsharp_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
            vsharp_mem = (8.15059 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 1.0839) # DONE
            mn_vsharp = create_node(
                interface=qsmjl.VsharpInterface(num_threads=vsharp_threads),
                iterfield=['frequency', 'mask'],
                name='qsmjl_vsharp',
                n_procs=vsharp_threads,
                mem_gb=vsharp_mem,
                is_map=use_maps
            )
            wf.connect([
                (n_phase_normalized, mn_vsharp, [('phase_normalized', 'frequency')]),
                (n_inputs, mn_vsharp, [('mask', 'mask')]),
                (n_inputs, mn_vsharp, [('vsz', 'vsz')]),
                (mn_vsharp, mn_bf, [('tissue_frequency', 'tissue_frequency')]),
                (mn_vsharp, mn_bf, [('vsharp_mask', 'mask')]),
            ])
            mn_vsharp.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=run_args.slurm[0],
                pbs_account=run_args.pbs,
                slurm_partition=run_args.slurm[1],
                name="VSHARP",
                mem_gb=vsharp_mem,
                num_cpus=vsharp_threads
            )
        if run_args.bf_algorithm == 'pdf':
            pdf_threads = min(8, run_args.n_procs) if run_args.multiproc else 8
            pdf_mem = 9.23275 * (np.prod(dimensions_phase) * 8) / (1024 ** 3) + 1.46 # DONE
            mn_pdf = create_node(
                interface=qsmjl.PdfInterface(num_threads=pdf_threads),
                iterfield=['frequency', 'mask'],
                name='qsmjl_pdf',
                n_procs=pdf_threads,
                mem_gb=pdf_mem,
                is_map=use_maps
            )
            wf.connect([
                (n_phase_normalized, mn_pdf, [('phase_normalized', 'frequency')]),
                (n_inputs, mn_pdf, [('mask', 'mask')]),
                (n_inputs, mn_pdf, [('vsz', 'vsz')]),
                (mn_pdf, mn_bf, [('tissue_frequency', 'tissue_frequency')]),
                (n_inputs, mn_bf, [('mask', 'mask')]),
            ])
            mn_pdf.plugin_args = gen_plugin_args(
                plugin_args={ 'overwrite': True },
                slurm_account=run_args.slurm[0],
                pbs_account=run_args.pbs,
                slurm_partition=run_args.slurm[1],
                name="PDF",
                time="01:00:00",
                mem_gb=pdf_mem,
                num_cpus=pdf_threads
            )

    # === DIPOLE INVERSION ===
    if run_args.qsm_algorithm == 'nextqsm':
        nextqsm_threads = min(8, run_args.n_procs) if run_args.multiproc else 8
        nextqsm_mem = 69.64824 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 5.6689 # DONE
        mn_qsm = create_node(
            interface=nextqsm.NextqsmInterface(),
            name='nextqsm',
            iterfield=['phase', 'mask'],
            mem_gb=nextqsm_mem,
            n_procs=nextqsm_threads,
            is_map=use_maps
        )
        wf.connect([
            (n_phase_normalized, mn_qsm, [('phase_normalized', 'phase')]),
            (n_inputs, mn_qsm, [('mask', 'mask')]),
            (mn_qsm, n_outputs, [('qsm', 'qsm')]),
        ])
        mn_qsm.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="NEXTQSM",
            mem_gb=nextqsm_mem,
            num_cpus=nextqsm_threads
        )
    if run_args.qsm_algorithm == 'rts':
        rts_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
        rts_mem = (18.19 * (np.prod(dimensions_phase) * 3 / (1024 ** 3)) + 2) # DONE
        mn_qsm = create_node(
            interface=qsmjl.RtsQsmInterface(num_threads=rts_threads),
            name='qsmjl_rts',
            iterfield=['tissue_frequency', 'mask'],
            n_procs=rts_threads,
            mem_gb=rts_mem,
            terminal_output="file_split",
            is_map=use_maps
        )
        wf.connect([
            (mn_bf, mn_qsm, [('tissue_frequency', 'tissue_frequency')]),
            (mn_bf, mn_qsm, [('mask', 'mask')]),
            (n_inputs, mn_qsm, [('vsz', 'vsz')]),
            (n_inputs, mn_qsm, [('b0_direction', 'b0_direction')]),
            (mn_qsm, n_outputs, [('qsm', 'qsm')]),
        ])
        mn_qsm.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="RTS",
            mem_gb=rts_mem,
            num_cpus=rts_threads
        )
    if run_args.qsm_algorithm == 'tv':
        tv_threads = min(2, run_args.n_procs) if run_args.multiproc else 2
        tv_mem = 4.5365 * (np.prod(dimensions_phase) * 8 / (1024 ** 3)) + 2 # DONE
        mn_qsm = create_node(
            interface=qsmjl.TvQsmInterface(num_threads=tv_threads),
            name='qsmjl_tv',
            iterfield=['tissue_frequency', 'mask'],
            n_procs=tv_threads,
            mem_gb=tv_mem,
            terminal_output="file_split",
            is_map=use_maps
        )
        wf.connect([
            (mn_bf, mn_qsm, [('tissue_frequency', 'tissue_frequency')]),
            (mn_bf, mn_qsm, [('mask', 'mask')]),
            (n_inputs, mn_qsm, [('vsz', 'vsz')]),
            (n_inputs, mn_qsm, [('b0_direction', 'b0_direction')]),
            (mn_qsm, n_outputs, [('qsm', 'qsm')]),
        ])
        mn_qsm.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name="TV",
            time="01:00:00",
            mem_gb=tv_mem,
            num_cpus=tv_threads
        )

    if run_args.qsm_algorithm == 'tgv':
        tgv_threads = min(4, run_args.n_procs)
        tgv_mem = 50.9915 * (np.prod(dimensions_phase) * 8 / (1024**3)) + 1.2 # DONE
        mn_qsm = create_node(
            interface=tgvjl.TGVQSMJlInterface(
                erosions=qsm_erosions,
                alpha=run_args.tgv_alphas,
                iterations=run_args.tgv_iterations,
                num_threads=tgv_threads
            ),
            name='tgv',
            iterfield=['phase', 'TE', 'mask'],
            is_map=use_maps,
            mem_gb=tgv_mem,
            n_procs=tgv_threads
        )
        mn_qsm.plugin_args = gen_plugin_args(
            plugin_args={ 'overwrite': True },
            slurm_account=run_args.slurm[0],
            pbs_account=run_args.pbs,
            slurm_partition=run_args.slurm[1],
            name='TGV',
            time="01:00:00",
            mem_gb=tgv_mem,
            num_cpus=tgv_threads
        )
        wf.connect([
            (n_inputs, mn_qsm, [('mask', 'mask')]),
            (n_inputs, mn_qsm, [('B0', 'B0')]),
            (mn_qsm, n_outputs, [('qsm', 'qsm')]),
        ])
        if run_args.combine_phase:
            mn_qsm.inputs.TE = 0.005
            wf.connect([
                (n_phase_normalized, mn_qsm, [('phase_normalized', 'phase')])
            ])
        else:
            wf.connect([
                (n_inputs, mn_qsm, [('phase', 'phase')]),
                (n_inputs, mn_qsm, [('TE', 'TE')])
            ])

    
    return wf




"""
Created on Sun Aug  3 11:46:42 2014

@author: epracht
modified by Steffen.Bollmann@cai.uq.edu.au
"""

from __future__ import division
from nipype.interfaces.base import CommandLine, traits, TraitedSpec, File, CommandLineInputSpec, InputMultiPath
from nipype.interfaces.base.traits_extension import isdefined
from qsmxt.scripts.qsmxt_functions import extend_fname
import os, shutil

THREAD_CONTROL_VARIABLE = "OMP_NUM_THREADS"


class TGVQSMInputSpec(CommandLineInputSpec):
    phase = File(exists=True, desc='Phase image', mandatory=True, argstr="-p %s")
    mask = InputMultiPath(exists=True, desc='Image mask', mandatory=True, argstr="-m %s")
    num_threads = traits.Int(-1, usedefault=True, nohash=True, desc="Number of threads to use, by default $NCPUS")
    TE = traits.Float(desc='Echo Time [sec]', mandatory=True, argstr="-t %f")
    B0 = traits.Float(desc='Field Strength [Tesla]', mandatory=True, argstr="-f %f")
    extra_arguments = traits.String(desc='Add extra arguments. E.G. --ignore-orientation --no-resampling will ignore orientation of files and do no resampling (for cases where resampling in tgv_qsm fails)',
                               argstr="%s")
    # Only support of one alpha here!
    alpha = traits.List([0.0015, 0.0005], minlen=2, maxlen=2, desc='Regularisation alphas', usedefault=True,
                        argstr="--alpha %s")
    # We only support one iteration - easier to handle in nipype
    iterations = traits.Int(1000, desc='Number of iterations to perform', usedefault=True, argstr="-i %d")
    erosions = traits.Int(5, desc='Number of mask erosions', usedefault=True, argstr="-e %d")
    out_suffix = traits.String("_tgv", desc='Suffix for output files. Will be followed by 000 (reason - see CLI)',
                               usedefault=True, argstr="-o %s")


class TGVQSMOutputSpec(TraitedSpec):
    qsm = File(desc='Computed susceptibility map')


class TGVQSMInterface(CommandLine):
    input_spec = TGVQSMInputSpec
    output_spec = TGVQSMOutputSpec

    # We use here an interface to the CLI utility
    _cmd = "tgv_qsm"

    def __init__(self, **inputs):
        super(TGVQSMInterface, self).__init__(**inputs)
        self.inputs.on_trait_change(self._num_threads_update, 'num_threads')

        if not isdefined(self.inputs.num_threads):
            self.inputs.num_threads = self._num_threads
        else:
            self._num_threads_update()

    def _list_outputs(self):
        outputs = self.output_spec().get()
        
        # TGV-QSM doesn't output files in the current directory for some reason, so we should move it
        outfile_original = extend_fname(self.inputs.phase, "_tgv_000", ext="nii.gz")
        outfile_final = os.path.abspath(os.path.split(outfile_original)[1]).replace("_000.nii.gz", ".nii.gz")
        if not os.path.exists(outfile_final):
            shutil.move(outfile_original, outfile_final)
        
        outputs['qsm'] = outfile_final
        return outputs

    def _num_threads_update(self):
        self._num_threads = self.inputs.num_threads
        if self.inputs.num_threads == -1:
            self.inputs.environ.update({THREAD_CONTROL_VARIABLE: '$NCPUS'})
        else:
            self.inputs.environ.update({THREAD_CONTROL_VARIABLE: '%s' % self.inputs.num_threads})



import os
import glob
import datetime
import signal
import fnmatch
from random import randint
from time import sleep

import osfclient
import webdav3.client
import nibabel as nib
import numpy as np

from matplotlib import pyplot as plt

from qsmxt.cli.main import *
from qsmxt.scripts.sys_cmd import sys_cmd
from qsmxt.scripts.logger import LogLevel, make_logger

def create_logger(log_dir):
    os.makedirs(log_dir, exist_ok=True)
    return make_logger(
        logpath=os.path.join(log_dir, f"log_{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}.txt"),
        printlevel=LogLevel.DEBUG,
        writelevel=LogLevel.INFO,
        warnlevel=LogLevel.WARNING,
        errorlevel=LogLevel.ERROR
    )

def write_to_file(path, text, mode='a', end="\n"):
    logger = make_logger()
    logger.log(LogLevel.DEBUG.value, f"Writing to {path}: {text}")
    with open(path, mode) as filehandle:
        filehandle.write(text + end)

def upload_png(png_path):
    logger = make_logger()
    logger.log(LogLevel.INFO.value, f"Uploading png {png_path}")
    png_url = sys_cmd(f"images-upload-cli --no-clipboard --hosting freeimage {png_path}").strip()
    return png_url

def webdav_connect():
    logger = make_logger()
    logger.log(LogLevel.INFO.value, "Retrieving WEBDAV details...")
    try:
        webdav_login = os.environ['WEBDAV_LOGIN']
        webdav_password = os.environ['WEBDAV_PASSWORD']
    except KeyError as e:
        logger.log(LogLevel.ERROR.value, f"Could not connect to WEBDAV - missing WEBDAV_LOGIN and/or WEBDAV_PASSWORD")
        raise e

    logger.log(LogLevel.INFO.value, f"Establishing WEBDAV connection...")
    try:
        client = webdav3.client.Client({
            'webdav_hostname': f"https://cloud.rdm.uq.edu.au/remote.php/dav/files/{webdav_login}/",
            'webdav_login':    webdav_login,
            'webdav_password': webdav_password,
            'webdav_timeout': 120
        })
    except Exception as e:
        logger.log(LogLevel.ERROR.value, f"Could not connect to WEBDAV - connection error!")
        raise e

    logger.log(LogLevel.INFO.value, "Connection successful!")
    return client

def upload_to_rdm(local_path, remote_path):
    logger = make_logger()
    client = webdav_connect()
    exception = None

    # Signal handler for the alarm
    def handler(signum, frame):
        raise TimeoutError("Upload took too long!")

    # Set the signal handler for the alarm signal
    signal.signal(signal.SIGALRM, handler)

    for i in range(5):
        logger.log(LogLevel.INFO.value, f"Uploading {local_path} to {remote_path}")
        try:
            # Set an alarm for 10 minutes
            signal.alarm(10 * 60)

            client.upload_sync(
                remote_path=remote_path,
                local_path=local_path
            )
            
            # Cancel the alarm if the upload is successful
            signal.alarm(0)

        except webdav3.exceptions.ConnectionException as e:
            logger.log(LogLevel.WARNING.value, f"Connection failed! {e}")
            sleeptime = randint(120, 300)
            logger.log(LogLevel.INFO.value, f"Sleeping for {sleeptime} seconds...")
            sleep(sleeptime)
            exception = e
            continue
        except TimeoutError as e:
            logger.log(LogLevel.WARNING.value, f"{e}")
            return
        break
    else:
        logger.log(LogLevel.ERROR.value, f"Upload failed after 5 attempts!")
        raise exception


def download_from_rdm(remote_path, local_path):
    logger = make_logger()
    client = webdav_connect()
    exception = None
    for i in range(5):
        logger.log(LogLevel.INFO.value, f"Downloading {remote_path} to {local_path}")
        try:
            client.download_sync(
                remote_path=remote_path,
                local_path=local_path,
            )
        except webdav3.exceptions.ConnectionException as e:
            logger.log(LogLevel.WARNING.value, f"Connection failed! {e}")
            sleeptime = randint(120, 300)
            logger.log(LogLevel.INFO.value, f"Sleeping for {sleeptime} seconds...")
            sleep(sleeptime)
            exception = e
            continue
        break
    else:
        logger.log(LogLevel.ERROR.value, f"Download failed after 5 attempts!")
        raise exception

def download_from_osf(project, local_path):
    logger = make_logger()
    try:
        osf_token = os.environ.get('OSF_TOKEN', '')
        osf_user = os.environ.get('OSF_USER', '') or os.environ.get('OSF_USERNAME', '')
        osf_pass = os.environ.get('OSF_PASS', '') or os.environ.get('OSF_PASSWORD', '')
    except KeyError as e:
        logger.log(LogLevel.ERROR.value, f"Cannot connect to OSF - missing environment variable/s! Need OSF_TOKEN, OSF_USER and OSF_PASS.")
        raise e
    
    if any(len(x) == 0 for x in [osf_token, osf_user, osf_pass]):
        if len(osf_token) == 0:
            logger.log(LogLevel.ERROR.value, f"OSF_TOKEN length is zero")
        if len(osf_user) == 0:
            logger.log(LogLevel.ERROR.value, f"OSF_USER length is zero")
        if len(osf_pass) == 0:
            logger.log(LogLevel.ERROR.value, f"OSF_PASS length is zero")
        raise ValueError("OSF_TOKEN, OSF_USER and/or OSF_PASS not initialised properly!")
    
    logger.log(LogLevel.INFO.value, "Connecting to OSF...")
    osf = osfclient.OSF(username=osf_user, password=osf_pass, token=osf_token)
    osf_project = osf.project(project)
    osf_file = list(osf_project.storage().files)[0]

    logger.log(LogLevel.INFO.value, f"Downloading from {project} to {local_path}")
    with open(local_path, 'wb') as fpr:
        osf_file.write_to(fpr)

def find_files(directory_pattern, search_pattern):
    matching_files = []

    for directory in glob.glob(directory_pattern):    
        for root, dirs, files in os.walk(os.path.abspath(directory)):
            for filename in fnmatch.filter(files, search_pattern):
                matching_files.append(os.path.join(root, filename))
    
    return matching_files

def compress_folder(folder, result_id):
    logger = make_logger()
    if os.environ.get('BRANCH'):
        results_tar = f"{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}_{os.environ['BRANCH']}_{result_id}.tar"
    else:
        results_tar = f"{str(datetime.datetime.now()).replace(':', '-').replace(' ', '_').replace('.', '')}_{result_id}.tar"

    logger.log(LogLevel.INFO.value, f"Compressing folder {folder} with suffix '{result_id}'")
    sys_cmd(f"tar -cf {results_tar} {folder}")

    return results_tar

def display_nii(
    nii_path=None, data=None, dim=0, title=None, slc=None, dpi=96, size=None, out_png=None, final_fig=True, title_fontsize=12,
    colorbar=False, cbar_label=None, cbar_orientation='vertical', cbar_nbins=None, cbar_fontsize=None, cbar_label_fontsize=8,
    **imshow_args
):
    data = data if data is not None else nib.load(nii_path).get_fdata()
    slc = slc or int(data.shape[0]/2)
    if dim == 0: slc_data = data[slc,:,:]
    if dim == 1: slc_data = data[:,slc,:]
    if dim == 2: slc_data = data[:,:,slc]
    if size:
        plt.figure(figsize=(size[0]/dpi, size[1]/dpi), dpi=dpi)
    else:
        plt.figure(dpi=dpi)
    plt.axis('off')
    plt.imshow(np.rot90(slc_data), **imshow_args)
    if colorbar:
        cbar = plt.colorbar(orientation=cbar_orientation, fraction=0.037, pad=0.04)
        if cbar_fontsize:
            cbar.ax.tick_params(labelsize=cbar_fontsize)
        if cbar_nbins:
            cbar.ax.locator_params(nbins=cbar_nbins)
        if cbar_label:
            if cbar_orientation == 'horizontal':
                cbar.ax.set_xlabel(cbar_label, fontsize=cbar_label_fontsize)
            else:
                cbar.ax.set_ylabel(cbar_label, fontsize=cbar_label_fontsize, rotation=90)
    if title:
        plt.title(title, fontsize=title_fontsize)
    if final_fig:
        if out_png:
            plt.savefig(out_png, bbox_inches='tight')
            return out_png
        else:
            plt.show()




import subprocess


def sys_cmd(cmd, print_output=True, print_command=True, raise_exception=False):
    if print_command:
        print(cmd)

    process = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    results = process.stdout.decode('UTF-8')[:-1] + '\n' + process.stderr.decode('UTF-8')[:-1]

    if print_output:
        print(results, end="")

    if raise_exception and process.returncode != 0:
        raise subprocess.CalledProcessError(process.returncode, cmd, output=process.stderr.decode('UTF-8')[:-1])

    return results




from nipype.interfaces.base import BaseInterface, BaseInterfaceInputSpec, TraitedSpec, File, Directory, traits, DynamicTraitedSpec, Undefined, isdefined
import shutil
import os

class DynamicCopyFilesInputSpec(DynamicTraitedSpec, BaseInterfaceInputSpec):
    output_map = traits.Dict(mandatory=True, desc="Dictionary mapping inputs to their target output paths")
    parameterization = traits.Bool(True, usedefault=True, desc="Store output in parametrized structure")
    remove_dest_dir = traits.Bool(False, usedefault=True, desc="Remove destination directory when copying dirs")

class DynamicCopyFilesOutputSpec(TraitedSpec):
    out_files = traits.List(File(exists=True), desc="List of copied output files")
    out_dirs = traits.List(Directory(exists=True), desc="List of copied output directories")

class DynamicCopyFiles(BaseInterface):
    input_spec = DynamicCopyFilesInputSpec
    output_spec = DynamicCopyFilesOutputSpec

    def __init__(self, infields=None, **kwargs):
        super().__init__(**kwargs)
        self._results = {}
        if infields:
            undefined_traits = {}
            for key in infields:
                self.inputs.add_trait(key, traits.Any)
                undefined_traits[key] = Undefined
            self.inputs.trait_set(trait_change_notify=False, **undefined_traits)

    def _run_interface(self, runtime):
        output_map = self.inputs.output_map
        copied_files = []
        copied_dirs = []

        for name in self.inputs.copyable_trait_names():
            if name in ['output_map', 'parameterization', 'remove_dest_dir']:
                continue

            value = getattr(self.inputs, name)
            if not isdefined(value):
                continue
            
            if isinstance(value, str):
                value = [value]

            if isinstance(value, list) and len(value) == 1:
                value = value[0]

            if isinstance(value, str):
                if name in output_map:
                    dst = output_map[name]
                else:
                    dst = self._substitute_path(name, value)

                # Ensure the target directory exists
                os.makedirs(os.path.dirname(dst), exist_ok=True)

                if os.path.exists(value):
                    if os.path.isfile(value):
                        file_extension = '.' + '.'.join(os.path.split(value)[1].split('.')[1:])
                        dst_with_extension = f"{dst}{file_extension}"
                        shutil.copy2(value, dst_with_extension)
                        copied_files.append(dst_with_extension)
                    elif os.path.isdir(value):
                        if os.path.exists(dst) and self.inputs.remove_dest_dir:
                            shutil.rmtree(dst)
                        shutil.copytree(value, dst)
                        copied_dirs.append(dst)
            elif isinstance(value, list):
                dst_dir = output_map[name] if name in output_map else os.path.join(output_map['base_directory'], name)
                os.makedirs(dst_dir, exist_ok=True)
                for src in value:
                    if os.path.isfile(src):
                        file_dst = os.path.join(dst_dir, os.path.basename(src))
                        shutil.copy2(src, file_dst)
                        copied_files.append(file_dst)
                    elif os.path.isdir(src):
                        dir_dst = os.path.join(dst_dir, os.path.basename(src))
                        if os.path.exists(dir_dst) and self.inputs.remove_dest_dir:
                            shutil.rmtree(dir_dst)
                        shutil.copytree(src, dir_dst)
                        copied_dirs.append(dir_dst)

        self._results['out_files'] = copied_files
        self._results['out_dirs'] = copied_dirs

        return runtime

    def _list_outputs(self):
        outputs = self._results
        return outputs

    def _substitute_path(self, name, path):
        return os.path.basename(path)  # Adjust this if you have specific path substitution logic
    
    


#!/usr/bin/env python

import nibabel as nib
import nilearn.image
import numpy as np
import argparse

def resample_to_reference(source_file, reference_file, output_file, interpolation='continuous'):
    # Load NIfTI files
    source_nii = nib.load(source_file)
    reference_nii = nib.load(reference_file)

    # Resample the source image to match the reference image
    resampled_nii = nilearn.image.resample_img(
        source_nii, 
        target_affine=reference_nii.affine, 
        target_shape=np.array(reference_nii.header.get_data_shape()),
        interpolation=interpolation
    )
    
    # Save the resampled image
    nib.save(resampled_nii, output_file)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Resample a NIfTI image to match the dimensions, resolution, and orientation of a reference image.")
    parser.add_argument("source_file", help="Path to the source NIfTI file to be resampled.")
    parser.add_argument("reference_file", help="Path to the reference NIfTI file.")
    parser.add_argument("output_file", help="Path to save the resampled NIfTI file.")
    parser.add_argument("--interpolation", choices=['nearest', 'continuous'], default='continuous', 
                        help="Type of interpolation for resampling ('nearest' or 'continuous'). Default is 'continuous'.")
    args = parser.parse_args()

    resample_to_reference(args.source_file, args.reference_file, args.output_file, args.interpolation)




import os
from nipype.interfaces.base import CommandLine, TraitedSpec, traits, File, Directory, CommandLineInputSpec
from nipype.utils.filemanip import split_filename

class Nii2DcmInputSpec(CommandLineInputSpec):
    dicom_type = traits.Str(
        "MR",
        usedefault=True,
        argstr="--dicom_type %s",
    )
    ref_dicom = File(
        exists=True,
        mandatory=False,
        argstr="--ref_dicom %s"
    )
    centered = traits.Bool(
        argstr="--centered"
    )
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0
    )
    out_dir = traits.Str(
        argstr="%s",
        name_source=['in_file'],
        name_template='%s_dcm',
        position=1
    )


class Nii2DcmOutputSpec(TraitedSpec):
    out_dir = traits.Directory(exists=True)


class Nii2DcmInterface(CommandLine):
    input_spec = Nii2DcmInputSpec
    output_spec = Nii2DcmOutputSpec
    _cmd = "git init && nii2dcm --dicom_type MR"

    def _format_arg(self, name, spec, value):
        if name == 'out_dir':
            path, base, _ = split_filename(self.inputs.in_file)
            path = os.path.join(os.path.abspath(path), f"{base}_dcm")
            os.makedirs(path, exist_ok=True)
            return path
        return super(Nii2DcmInterface, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        out_dir = self._format_arg('out_dir', self.inputs.trait('out_dir'), self.inputs.out_dir)
        outputs['out_dir'] = out_dir
        return outputs




#!/usr/bin/env python3
import os
from qsmxt.scripts.qsmxt_functions import extend_fname
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, traits, File

def erosion(in_file, num_erosions=1):
    if num_erosions == 0: return in_file

    import os
    import nibabel as nib
    from scipy.ndimage import binary_erosion
    
    # load data
    nii = nib.load(in_file)
    data = nii.get_fdata()

    # erosions
    for i in range(num_erosions):
        data = binary_erosion(data).astype(int)

    # write to file
    out_file = extend_fname(in_file, "_ero", out_dir=os.getcwd())
    nib.save(nib.Nifti1Image(dataobj=data, header=nii.header, affine=nii.affine), out_file)

    return out_file


class ErosionInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True)
    num_erosions = traits.Int(default=1)


class ErosionOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class ErosionInterface(SimpleInterface):
    input_spec = ErosionInputSpec
    output_spec = ErosionOutputSpec

    def _run_interface(self, runtime):
        out_file = erosion(self.inputs.in_file, self.inputs.num_erosions)
        self._results['out_file'] = out_file
        return runtime


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'in_file',
        required=True,
        type=str
    )

    parser.add_argument(
        '--num_erosions',
        required=False,
        default=1,
        type=int
    )

    args = parser.parse_args()
    mask_files = erosion(args.in_file, args.num_erosions)




import os
import nibabel as nib
import numpy as np
import nilearn.image
import warnings
from qsmxt.scripts.qsmxt_functions import extend_fname
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, traits

def resample_to_axial(mag_nii=None, pha_nii=None, mask_nii=None):
    # calculate base affine
    nii = mag_nii or pha_nii or mask_nii
    voxel_size = np.array(nii.header.get_zooms())
    resolution = np.array(nii.header.get_data_shape())
    origin = np.array(voxel_size * resolution / 2)
    base_affine = np.eye(4)
    np.fill_diagonal(base_affine, voxel_size * np.sign(np.diag(nii.affine))[:3])
    base_affine[3,3] = 1
    base_affine[:3,3] = origin * -np.sign(np.diag(nii.affine)[:3])
    base_affine = base_affine[:3,:3]

    mag_rot_nii = None
    pha_rot_nii = None
    mask_rot_nii = None

    # compute real and imaginary components from magnitude and phase
    if mag_nii and pha_nii:
        pha = pha_nii.get_fdata()
        mag = mag_nii.get_fdata()
        real = mag * np.cos(pha)
        imag = mag * np.sin(pha)
        cplx_header = mag_nii.header.copy()
        cplx_header.set_data_dtype(np.float32)
        real_nii = nib.Nifti1Image(real, affine=pha_nii.affine, header=cplx_header)
        imag_nii = nib.Nifti1Image(imag, affine=pha_nii.affine, header=cplx_header)

        # resample real and imaginary to base affine
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            real_rot_nii = nilearn.image.resample_img(real_nii, target_affine=base_affine, target_shape=None, interpolation='continuous')
            imag_rot_nii = nilearn.image.resample_img(imag_nii, target_affine=base_affine, target_shape=None, interpolation='continuous')
            mask_rot_nii = nilearn.image.resample_img(mask_nii, target_affine=base_affine, target_shape=None, interpolation='nearest') if mask_nii else None

        # convert real and imaginary to magnitude and phase
        real_rot = real_rot_nii.get_fdata()
        imag_rot = imag_rot_nii.get_fdata()
        mag_rot = np.array(np.round(np.hypot(real_rot, imag_rot, dtype=np.float32), 0), dtype=np.int32)
        pha_rot = np.arctan2(imag_rot, real_rot, dtype=np.float32)

        # add noise to zero values
        mask = pha_rot == 0
        if mask.sum() / mask.size >= 0.1:
            np.random.seed()
            noise = np.random.uniform(-np.pi, np.pi, pha_rot.shape)
            pha_rot[mask] = noise[mask]

        # create nifti objects
        mag_rot_nii = nib.Nifti1Image(mag_rot, affine=real_rot_nii.affine, header=mag_nii.header)
        pha_rot_nii = nib.Nifti1Image(pha_rot, affine=real_rot_nii.affine, header=pha_nii.header)
    elif mag_nii:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mag_rot_nii = nilearn.image.resample_img(mag_nii, target_affine=base_affine, target_shape=None, interpolation='continuous')
    if mask_nii:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mask_rot_nii = nilearn.image.resample_img(mask_nii, target_affine=base_affine, target_shape=None, interpolation='nearest')

    return mag_rot_nii, pha_rot_nii, mask_rot_nii

def resample_files(mag_file=None, pha_file=None, mask_file=None, obliquity_threshold=None):
    # load data
    #print(f"Loading mag={os.path.split(mag_file)[1]}...")
    mag_nii = nib.load(mag_file) if mag_file else None
    #print(f"Loading pha={os.path.split(pha_file)[1]}...")
    pha_nii = nib.load(pha_file) if pha_file else None
    #if mask_file:
    #    print(f"Loading mask={os.path.split(mask_file)[1]}...")
    mask_nii = nib.load(mask_file) if mask_file else None        

    # check obliquity
    nii = mag_nii or pha_nii or mask_nii
    obliquity = np.rad2deg(nib.affines.obliquity(nii.affine))
    obliquity_norm = np.linalg.norm(obliquity)
    if obliquity_threshold and obliquity_norm < obliquity_threshold:
        #print(f"Obliquity = {obliquity}; norm = {obliquity_norm} < {obliquity_threshold}; no resampling needed.")
        return mag_file, pha_file, mask_file
    #print(f"Obliquity = {obliquity}; norm = {obliquity_norm} >= {obliquity_threshold}; resampling will commence.")

    # resample
    mag_rot_nii, pha_rot_nii, mask_rot_nii = resample_to_axial(mag_nii, pha_nii, mask_nii)
    
    # save results
    mag_resampled_fname = None
    pha_resampled_fname = None
    mask_resampled_fname = None
    if mag_rot_nii:
        mag_resampled_fname = extend_fname(mag_file, "_resampled", out_dir=os.getcwd())
        nib.save(mag_rot_nii, mag_resampled_fname)
    if pha_rot_nii:
        pha_resampled_fname = extend_fname(pha_file, "_resampled", out_dir=os.getcwd())
        nib.save(pha_rot_nii, pha_resampled_fname)
    if mask_rot_nii:
        mask_resampled_fname = extend_fname(mask_file, "_resampled", out_dir=os.getcwd())
        nib.save(mask_rot_nii, mask_resampled_fname)

    return mag_resampled_fname, pha_resampled_fname, mask_resampled_fname


def resample_like(in_file, in_like, interpolation='continuous'):
    in_nii = nib.load(in_file)
    in_like_nii = nib.load(in_like)
    if np.array_equal(in_nii.affine, in_like_nii.affine):
        return in_file
    in_nii_resampled = nilearn.image.resample_img(in_nii, target_affine=in_like_nii.affine, target_shape=np.array(in_like_nii.header.get_data_shape()), interpolation=interpolation)
    in_resampled_fname = extend_fname(in_file, "_resampled", out_dir=os.getcwd())
    nib.save(in_nii_resampled, in_resampled_fname)
    return in_resampled_fname


class AxialSamplingInputSpec(BaseInterfaceInputSpec):
    magnitude = File(mandatory=False, exists=True)
    phase = File(mandatory=False, exists=True)
    mask = File(mandatory=False, exists=True)
    obliquity_threshold = traits.Float(mandatory=False)


class AxialSamplingOutputSpec(TraitedSpec):
    magnitude = File(exists=False)
    phase = File(exists=False)
    mask = File(mandatory=False)


class AxialSamplingInterface(SimpleInterface):
    input_spec = AxialSamplingInputSpec
    output_spec = AxialSamplingOutputSpec

    def _run_interface(self, runtime):
        magnitude, phase, mask = resample_files(
            mag_file=self.inputs.magnitude,
            pha_file=self.inputs.phase,
            mask_file=self.inputs.mask,
            obliquity_threshold=self.inputs.obliquity_threshold
        )
        if magnitude: self._results['magnitude'] = magnitude
        if phase: self._results['phase'] = phase
        if mask: self._results['mask'] = mask
        
        return runtime


class ResampleLikeInputSpec(BaseInterfaceInputSpec):
    in_file = File(mandatory=True, exists=True)
    in_like = File(mandatory=True, exists=True)


class ResampleLikeOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class ResampleLikeInterface(SimpleInterface):
    input_spec = ResampleLikeInputSpec
    output_spec = ResampleLikeOutputSpec

    def _run_interface(self, runtime):
        out_file = resample_like(self.inputs.in_file, self.inputs.in_like)
        self._results['out_file'] = out_file
        return runtime




import os
from nipype.interfaces.base import CommandLine, CommandLineInputSpec, traits
from nipype.interfaces.base.traits_extension import isdefined

class CommandLineInputSpecJulia(CommandLineInputSpec):
    num_threads = traits.Int(-1, usedefault=True, desc="Number of threads to use, by default $NCPUS")
    def __init__(self, **inputs): super(CommandLineInputSpecJulia, self).__init__(**inputs)

class CommandLineJulia(CommandLine):
    def __init__(self, **inputs):
        super(CommandLineJulia, self).__init__(**inputs)
        self.inputs.on_trait_change(self._num_threads_update, 'num_threads')

        if not isdefined(self.inputs.num_threads):
            self.inputs.num_threads = self._num_threads
        else:
            self._num_threads_update()

    def _num_threads_update(self):
        self._num_threads = self.inputs.num_threads
        if self.inputs.num_threads == -1:
            cpu_count = max(4, int(os.environ["NCPUS"]) if "NCPUS" in os.environ else int(os.cpu_count()))
            self.inputs.environ.update({ "JULIA_NUM_THREADS" : str(cpu_count), "JULIA_CPU_THREADS" : str(cpu_count) })
        else:
            self.inputs.environ.update({ "JULIA_NUM_THREADS" : f"{self.inputs.num_threads}", "JULIA_CPU_THREADS" : f"{self.inputs.num_threads}" })




#!/usr/bin/env python3
from qsmxt.scripts.qsmxt_functions import extend_fname
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, InputMultiPath

def combine_magnitude(magnitude_files, out_file=None):
    import os
    import nibabel as nib
    import numpy as np
    
    # load data
    nii = nib.load(magnitude_files[0])
    data = np.array([nib.load(magnitude_files[i]).get_fdata() for i in range(len(magnitude_files))])
    data = np.linalg.norm(data, axis=0)

    # write to file
    out_file = out_file or extend_fname(magnitude_files[0], "_combined", out_dir=os.getcwd())
    nib.save(nib.Nifti1Image(dataobj=data, header=nii.header, affine=nii.affine), out_file)

    return out_file


class CombineMagnitudeInputSpec(BaseInterfaceInputSpec):
    magnitude = InputMultiPath(mandatory=True, exists=True)


class CombineMagnitudeOutputSpec(TraitedSpec):
    magnitude_combined = File(exists=True)


class CombineMagnitudeInterface(SimpleInterface):
    input_spec = CombineMagnitudeInputSpec
    output_spec = CombineMagnitudeOutputSpec

    def _run_interface(self, runtime):
        magnitude_combined = combine_magnitude(self.inputs.magnitude)
        self._results['magnitude_combined'] = magnitude_combined
        return runtime


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'magnitude',
        required=True,
        nargs='+',
        type=str
    )

    parser.add_argument(
        'out_file',
        required=False,
        default=None,
        type=str
    )

    args = parser.parse_args()
    mask_files = combine_magnitude(args.magnitude, args.out_file)




from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from nipype.interfaces.base import TraitedSpec, File
from qsmxt.scripts import qsmxt_functions
import os

class MakeHomogeneousInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(MakeHomogeneousInputSpec, self).__init__(**inputs)
    magnitude = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0
    )
    magnitude_corrected = File(
        argstr="%s",
        name_source=['magnitude'],
        name_template='%s_makehomogeneous.nii',
        position=1
    )


class MakeHomogeneousOutputSpec(TraitedSpec):
    magnitude_corrected = File()


class MakeHomogeneousInterface(CommandLineJulia):
    def __init__(self, **inputs): super(MakeHomogeneousInterface, self).__init__(**inputs)
    input_spec = MakeHomogeneousInputSpec
    output_spec = MakeHomogeneousOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "makehomogeneous.jl")



#!/usr/bin/env python3

import argparse
import os
import shutil
import sys
import nibabel as nib
import numpy as np

from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from nipype.interfaces.base import  traits, TraitedSpec, File, InputMultiPath, OutputMultiPath
from qsmxt.scripts.qsmxt_functions import extend_fname, get_qsmxt_dir
    
def merge_multi_echo(in_paths, out_path):
    if len(in_paths) == 1: return in_paths[0]
    image4d = np.stack([nib.load(f).get_fdata() for f in in_paths], -1)
    sample_nii = nib.load(in_paths[0])
    nib.save(nib.nifti1.Nifti1Image(image4d, affine=sample_nii.affine, header=sample_nii.header), out_path)
    return out_path

def split_multi_echo(in_path, out_paths):
    image4d = nib.load(in_path).get_fdata()
    if len(image4d.shape) != 4 and len(out_paths) == 1:
        shutil.copy(in_path, out_paths[0])
        return out_paths
    sample_nii = nib.load(in_path)

    for i, out_path in enumerate(out_paths):
        echo_data = image4d[..., i]
        echo_nii = nib.nifti1.Nifti1Image(echo_data, affine=sample_nii.affine, header=sample_nii.header)
        nib.save(echo_nii, out_path)

    return out_paths

def wrap_phase(phase_path):
    phase_nii = nib.load(phase_path)
    phase = phase_nii.get_fdata()
    phase_wrapped = (phase + np.pi) % (2 * np.pi) - np.pi
    phase_wrapped_path = extend_fname(phase_path, "_wrapped", ext="nii", out_dir=os.getcwd())
    nib.save(img=nib.Nifti1Image(dataobj=phase_wrapped, affine=phase_nii.affine, header=phase_nii.header), filename=phase_wrapped_path)
    return phase_wrapped_path

class RomeoB0InputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(RomeoB0InputSpec, self).__init__(**inputs)
    # required inputs
    phase = InputMultiPath(mandatory=True, exists=True)
    magnitude = InputMultiPath(mandatory=False, exists=True)
    TEs = traits.ListFloat(mandatory=False, argstr="-t '[%s]'")
    TE = traits.Float(mandatory=False, argstr="-t '[%s]'")
    phase_offset_correction = traits.Enum("bipolar", "monopolar", usedefault=True, default="bipolar", argstr="--phase-offset-correction %s")
    
    # automatically filled
    combine_phase = File(exists=True, argstr="--phase %s", position=0)
    combine_mag = File(mandatory=False, exists=True, argstr="--mag %s", position=1)

class RomeoB0OutputSpec(TraitedSpec):
    frequency = File()
    phase_unwrapped = OutputMultiPath(File())
    #phase_unwrapped = File(exists=True)
    #phase_wrapped = File(exists=True)

class RomeoB0Interface(CommandLineJulia):
    def __init__(self, **inputs): super(RomeoB0Interface, self).__init__(**inputs)
    input_spec = RomeoB0InputSpec
    output_spec = RomeoB0OutputSpec
    _cmd = os.path.join(get_qsmxt_dir(), "scripts", "romeo_unwrapping.jl --no-rescale --compute-B0 B0.nii --individual-unwrapping --correct-global")

    def _format_arg(self, name, trait_spec, value):
        if name == 'TEs' or name == 'TE':
            if self.inputs.TEs is None and self.inputs.TE is None:
                raise ValueError("Either TEs or TE must be provided")
        return super(RomeoB0Interface, self)._format_arg(name, trait_spec, value)

    def _run_interface(self, runtime):
        num_phase_files = len(self.inputs.phase)
        self.inputs.phase_offset_correction = "bipolar" if num_phase_files >= 3 else "monopolar"

        if len(self.inputs.phase) > 1:
            self.inputs.combine_phase = merge_multi_echo(self.inputs.phase, os.path.join(os.getcwd(), "multi-echo-phase.nii"))
        else:
            self.inputs.combine_phase = self.inputs.phase[0]

        if len(self.inputs.magnitude) > 1:
            self.inputs.combine_mag = merge_multi_echo(self.inputs.magnitude, os.path.join(os.getcwd(), "multi-echo-mag.nii"))
        elif self.inputs.magnitude:
            self.inputs.combine_mag = self.inputs.magnitude[0]
        
        if self.inputs.TEs:
            self.inputs.TEs = [TE*1000 for TE in self.inputs.TEs]
        if self.inputs.TE:
            self.inputs.TE = self.inputs.TE*1000
        
        return super(RomeoB0Interface, self)._run_interface(runtime)
        
    def _list_outputs(self):
        outputs = self.output_spec().trait_get()

        # rename unwrapped.nii to suitable output name
        outputs['phase_unwrapped'] = split_multi_echo("unwrapped.nii", [extend_fname(f, "_romeo-unwrapped", ext="nii", out_dir=os.getcwd()) for f in self.inputs.phase])

        frequency_path = extend_fname(self.inputs.phase[0], "_B0", ext="nii", out_dir=os.getcwd())
        if os.path.exists('B0.nii'):
            os.rename('B0.nii', frequency_path)
        outputs['frequency'] = frequency_path

        return outputs


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process multi-echo phase and magnitude images to compute unwrapped phase and B0 map.")
    
    parser.add_argument("--phase", nargs='+', required=True, help="Path(s) to the phase image(s).")
    parser.add_argument("--magnitude", nargs='*', default=[], help="Path(s) to the magnitude image(s). Optional.")
    parser.add_argument("--TEs", nargs='*', type=float, default=None, help="Echo times in seconds for multi-echo sequences. Provide multiple values if using multiple echoes.")
    parser.add_argument("--TE", type=float, required=False, help="Echo time in seconds if using a single echo.")
    parser.add_argument("--phase-offset-correction", choices=["bipolar", "monopolar"], default="bipolar", help="Type of phase offset correction to apply.")
    
    args = parser.parse_args()

    try:
        romeo_interface = RomeoB0Interface(phase=args.phase,
                                           magnitude=args.magnitude if args.magnitude else None,
                                           TEs=args.TEs if args.TEs else None,
                                           phase_offset_correction=args.phase_offset_correction)
        romeo_interface.run()
        outputs = romeo_interface.aggregate_outputs()
        print(f"Processing complete. Unwrapped phase and B0 map generated.")
        print(f"Unwrapped Phase Path(s): {outputs.phase_unwrapped}")
        print(f"B0 Map Path: {outputs.frequency}")
        
    except Exception as e:
        sys.stderr.write(f"Error running RomeoB0Interface: {e}")
        sys.exit(1)



import os
import glob
import re

from nipype.interfaces.io import DataSink
from nipype.pipeline.engine import Workflow, Node
from nipype.interfaces.utility import IdentityInterface, Function
import nipype.interfaces.ants as ants

from qsmxt.scripts.qsmxt_functions import gen_plugin_args
from qsmxt.scripts.antsBuildTemplate import ANTSTemplateBuildSingleIterationWF
from qsmxt.scripts.logger import LogLevel, make_logger

def get_matching_files(bids_dir, subject, dtype="anat", suffixes=[], ext="nii*", session=None, run=None, part=None, acq=None):
    pattern = os.path.join(bids_dir, subject)
    if session:
        pattern = os.path.join(pattern, session)
    pattern = os.path.join(pattern, dtype) + os.path.sep
    if acq:
        pattern += f"*acq-{acq}_*"
    if run:
        pattern += f"*run-{run}_*"
    if part:
        pattern += f"*part-{part}_*"
    dir, fname = os.path.split(pattern)
    if suffixes:
        if fname:
            matching_files = [glob.glob(f"{pattern}_{suffix}.{ext}") for suffix in suffixes]
        else:
            matching_files = [glob.glob(os.path.join(dir, f"*{suffix}.{ext}")) for suffix in suffixes]
    else:
        matching_files = [glob.glob(f"{pattern}.{ext}")]
    return sorted([item for sublist in matching_files for item in sublist])

def init_template_workflow(args):
    logger = make_logger('main')
    logger.log(LogLevel.INFO.value, f"Creating QSM/GRE template workflow...")

    subjects = [
        os.path.split(path)[1]
        for path in glob.glob(os.path.join(args.bids_dir, "sub*"))
        if not args.subjects or os.path.split(path)[1] in args.subjects
    ]

    magnitude_files = []

    for subject in subjects:
        sessions = []
        if not glob.glob(os.path.join(args.bids_dir, subject, "ses*")):
            sessions = [None]
        else:
            sessions = [
                os.path.split(path)[1]
                for path in sorted(glob.glob(os.path.join(args.bids_dir, subject, "ses*")))
                if not args.sessions or os.path.split(path)[1] in args.sessions
            ]

        for session in sessions:
            phase_pattern = os.path.join(args.bids_dir, os.path.join(subject, session) if session else subject, "anat", f"sub-*_part-phase*.nii*")
            files = sorted(glob.glob(phase_pattern))

            if not files:
                logger.log(LogLevel.WARNING.value, f"No files found matching pattern: {phase_pattern}")
                continue

            run_details = {}

            # get all unique acquisition names
            acquisitions = sorted(list(set([
                re.search("_acq-([a-zA-Z0-9]+)_", path).group(1)
                for path in files
                if '_acq-' in path
            ])))

            # handle multiple runs associated with each acquisition name
            for acquisition in acquisitions:
                acquisition_runs = sorted(list(set([
                    re.search("_run-([a-zA-Z0-9]+)_", path).group(1)
                    for path in files
                    if acquisition in path
                    and '_run-' in path
                ])))
                run_details[acquisition] = acquisition_runs if len(acquisition_runs) else None

            # handle runs not associated with any acquisition name
            other_runs = sorted(list(set([
                re.search("_run-([a-zA-Z0-9]+)_", path).group(1)
                for path in files
                if '_acq-' not in path
                and '_run-' in path
            ])))
            if other_runs:
                run_details[None] = other_runs

            # handle the final case where there may be no run or acquisition identifier
            others = any(all(x not in path for x in ['_acq-', '_run-']) for path in files)
            if others:
                if None in run_details.keys(): run_details[None].append(None)
                run_details[None] = [None]

            if not run_details:
                continue

            for acq, runs in (run_details.items() if run_details else [(None, [None])]):
                for run in (runs if runs is not None else [None]):
                    run_magfiles = get_matching_files(args.bids_dir, subject=subject, session=session, acq=acq, run=run, dtype="anat", part="mag")
                    if run_magfiles:
                        magnitude_files.append(run_magfiles[0])

    if not magnitude_files:
        logger.log(LogLevel.ERROR.value, "No GRE magnitude images found! Template-building will not be possible.")
        return
    
    params_files = [path.replace('.nii.gz', '.nii').replace('.nii', '.json') for path in magnitude_files]

    wf = Workflow("template", base_dir=os.path.join(args.output_dir, "workflow", "template"))

    n_inputs = Node(
        IdentityInterface(
            fields=['magnitude', 'qsm', 'params']
        ),
        name='template_inputs'
    )
    n_inputs.inputs.magnitude = magnitude_files
    n_inputs.inputs.params = params_files
    n_inputs.inputs.sort_filelist = True

    n_outputs = Node(
        interface=IdentityInterface(
            fields=['initial_average', 'magnitude_template', 'qsm_template', 'transforms', 'qsms_transformed']
        ),
        name='template_outputs'
    )
    n_datasink = Node(
        interface=DataSink(base_directory=args.output_dir),
        name='template_datasink'
    )
    wf.connect([
        (n_outputs, n_datasink, [('initial_average', 'template.initial_average')]),
        (n_outputs, n_datasink, [('magnitude_template', 'template.magnitude_template')]),
        (n_outputs, n_datasink, [('qsm_template', 'template.qsm_template')]),
        (n_outputs, n_datasink, [('transforms', 'template.transforms')]),
        (n_outputs, n_datasink, [('qsms_transformed', 'template.qsms_transformed')])
    ])

    n_qsmdict = Node(
        interface=Function(
            input_names=['qsm'],
            output_names=['qsm_dict'],
            function=lambda qsm: [{'QSM': x} for x in qsm]
        ),
        name='func_create-qsm-dict'
    )
    wf.connect([
        (n_inputs, n_qsmdict, [('qsm', 'qsm')])
    ])

    # initial average
    initAvg = Node(
        interface=ants.AverageImages(),
        name='ants_average-images'
    )
    initAvg.inputs.dimension = 3
    initAvg.inputs.normalize = True
    wf.connect([
        (n_inputs, initAvg, [('magnitude', 'images')])
    ])

    # first iteration
    buildTemplateIteration1 = ANTSTemplateBuildSingleIterationWF('iteration01')
    wf.connect([
        (initAvg, buildTemplateIteration1, [('output_average_image', 'inputspec.fixed_image')]),
        (n_inputs, buildTemplateIteration1, [('magnitude', 'inputspec.images')]),
        (n_qsmdict, buildTemplateIteration1, [('qsm_dict', 'inputspec.ListOfPassiveImagesDictionaries')]),
    ])
    n_ants_threads = min(6, args.n_procs) if args.multiproc else 6
    n_ants_mem_gb = min(8, args.mem_avail) if args.multiproc else 8
    BeginANTS1 = buildTemplateIteration1.get_node("BeginANTS")
    BeginANTS1.plugin_args = gen_plugin_args(
        plugin_args={ 'overwrite': True },
        slurm_account=args.slurm[0],
        pbs_account=args.pbs,
        slurm_partition=args.slurm[1],
        name="ANTS",
        time="04:00:00",
        mem_gb=n_ants_mem_gb,
        num_cpus=n_ants_threads
    )
    BeginANTS1.environ = {
        'OMP_NUM_THREADS': str(n_ants_threads),
        'ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS': str(n_ants_threads)
    }

    # second iteration
    buildTemplateIteration2 = ANTSTemplateBuildSingleIterationWF('iteration02')
    wf.connect([
        (buildTemplateIteration1, buildTemplateIteration2, [('outputspec.template', 'inputspec.fixed_image')]),
        (n_inputs, buildTemplateIteration2, [('magnitude', 'inputspec.images')]),
        (n_qsmdict, buildTemplateIteration2, [('qsm_dict', 'inputspec.ListOfPassiveImagesDictionaries')])
    ])
    BeginANTS2 = buildTemplateIteration2.get_node("BeginANTS")
    BeginANTS2.plugin_args = gen_plugin_args(
        plugin_args={ 'overwrite': True },
        slurm_account=args.slurm[0],
        pbs_account=args.pbs,
        slurm_partition=args.slurm[1],
        name="ANTS",
        time="04:00:00",
        mem_gb=n_ants_mem_gb,
        num_cpus=n_ants_threads
    )
    BeginANTS2.environ = {
        'OMP_NUM_THREADS': str(n_ants_threads),
        'ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS': str(n_ants_threads)
    }

    # datasink
    wf.connect([
        (initAvg, n_outputs, [('output_average_image', 'initial_average')]),
        (buildTemplateIteration2, n_outputs, [('outputspec.template', 'magnitude_template')]),
        (buildTemplateIteration2, n_outputs, [('outputspec.passive_deformed_templates', 'qsm_template')]),
        (buildTemplateIteration2, n_outputs, [('outputspec.flattened_transforms', 'transforms')]),
        (buildTemplateIteration2, n_outputs, [('outputspec.wimtPassivedeformed', 'qsms_transformed')])
    ])

    return wf




from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from nipype.interfaces.base import  TraitedSpec, File
from qsmxt.scripts import qsmxt_functions
import os

## Laplacian wrapper
class LaplacianInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(LaplacianInputSpec, self).__init__(**inputs)
    phase = File(position=0, mandatory=True, exists=True, argstr='%s')
    phase_unwrapped = File(position=1, name_source=['phase'], name_template='%s_laplacian-unwrapped.nii.gz', argstr="%s")

class LaplacianOutputSpec(TraitedSpec):
    phase_unwrapped = File()

class LaplacianInterface(CommandLineJulia):
    def __init__(self, **inputs): super(LaplacianInterface, self).__init__(**inputs)
    input_spec = LaplacianInputSpec
    output_spec = LaplacianOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "mrt_laplacian_unwrapping.jl")
    



from qsmxt.interfaces.utils import CommandLineInputSpecJulia, CommandLineJulia
from nipype.interfaces.base import TraitedSpec, File, traits, InputMultiPath
from qsmxt.scripts import qsmxt_functions
import os


class ClearSwiInputSpec(CommandLineInputSpecJulia):
    def __init__(self, **inputs): super(ClearSwiInputSpec, self).__init__(**inputs)
    phase = InputMultiPath(
        exists=True,
        mandatory=True,
        argstr="--phase %s"
    )
    magnitude = InputMultiPath(
        exists=True,
        mandatory=True,
        argstr="--magnitude %s"
    )
    TEs = traits.ListFloat(
        mandatory=False,
        argstr="--TEs '[%s]'"
    )
    TE = traits.Float(
        mandatory=False,
        argstr="--TEs '[%s]'"
    )
    swi = File(
        argstr="--swi-out %s",
        name_source=['phase'],
        name_template='%s_swi.nii'
    )
    swi_mip = File(
        argstr="--mip-out %s",
        name_source=['phase'],
        name_template='%s_swi-mip.nii'
    )

class ClearSwiOutputSpec(TraitedSpec):
    swi = File()
    swi_mip = File()

class ClearSwiInterface(CommandLineJulia):
    def __init__(self, **inputs): super(ClearSwiInterface, self).__init__(**inputs)
    input_spec = ClearSwiInputSpec
    output_spec = ClearSwiOutputSpec
    _cmd = os.path.join(qsmxt_functions.get_qsmxt_dir(), "scripts", "mrt_clearswi.jl")

    def _format_arg(self, name, trait_spec, value):
        if name == 'TEs' or name == 'TE':
            if self.inputs.TEs is None and self.inputs.TE is None:
                raise ValueError("Either TEs or TE must be provided")
        return super(ClearSwiInterface, self)._format_arg(name, trait_spec, value)




#!/usr/bin/env python3

import os
import nibabel as nib
import numpy as np
from scipy.stats import norm
from scipy.ndimage import binary_fill_holes, binary_dilation, binary_erosion, gaussian_filter, binary_opening, convolve
from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, File, traits, InputMultiPath, OutputMultiPath
from qsmxt.scripts.qsmxt_functions import extend_fname
from skimage import filters

# === HELPER FUNCTIONS ===

def _histogram(image_histogram, normalize):
    bin_edges = np.histogram_bin_edges(image_histogram, bins='fd')
    hist, _ = np.histogram(image_histogram, bins=bin_edges)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.
    std = np.std(image_histogram)
    mean = np.mean(image_histogram)
    if normalize:
        hist = hist / np.sum(hist)
    return hist, bin_centers, mean, std

def _gaussian_threshold(image_histogram):
    hist, bin, mu, std = _histogram(image_histogram, True)
    normal_distribution = norm.pdf(bin, mu, std)
    difference = [normal_distribution[i] - hist[i] if hist[i]<normal_distribution[i] else 0 for i in range(len(hist)) ]
    maxpoint = max(range(len(difference)), key=difference.__getitem__)
    #threshold_percent = bin[maxpoint]/np.amax(image_histogram)*100
    return bin[maxpoint] #threshold_percent

def _clean_histogram(image_histogram):
    p_lower = np.percentile(image_histogram, 0.05) 
    p_upper = np.percentile(image_histogram, 99.5)
    image_histogram = image_histogram[np.logical_and(image_histogram > p_lower, image_histogram < p_upper)]
    return image_histogram

# === THRESHOLD-BASED MASKING FOR TWO-PASS AND SINGLE-PASS QSM ===
def threshold_masking(
        in_files,
        bet_masks=None,
        user_threshold=None,
        threshold_algorithm='gaussian',
        threshold_algorithm_factor=1.0,
        filling_algorithm='both',
        num_erosions=0,
        mask_suffix="_mask",
        fill_masks=False
    ):
    
    # sort input filepaths
    in_files = sorted(in_files)

    # load data
    all_niis = [nib.load(in_file) for in_file in in_files]
    all_float_data = [nii.get_fdata() for nii in all_niis]
    
    # calculate gaussian threshold if none given
    def get_threshold(data):
        if not user_threshold:
            image_histogram = np.array(data.flatten())
            if threshold_algorithm == 'gaussian':
                threshold = _gaussian_threshold(image_histogram)
            else:
                threshold = filters.threshold_otsu(image_histogram)
            threshold *= threshold_algorithm_factor
        elif round(user_threshold) == user_threshold: # user-defined absolute threshold
            threshold = user_threshold
        else: # user-defined percentage threshold
            data_range = np.max(np.array(all_float_data)) - np.min(np.array(all_float_data))
            threshold = np.min(np.array(all_float_data)) + (user_threshold * data_range)
        return threshold

    # do masking
    thresholds = [get_threshold(data) for data in all_float_data]
    masks = [np.array(all_float_data[i] > thresholds[i], dtype=int) for i in range(len(all_float_data))]
    masks = [fill_small_holes(mask) for mask in masks]
    if fill_masks:
        if filling_algorithm in ['gaussian', 'both']:
            masks = [fill_holes_smoothing(mask) for mask in masks]
        if filling_algorithm in ['morphological', 'both']:
            masks = [fill_holes_morphological(mask) for mask in masks]
        if num_erosions:
            masks = [np.array(binary_erosion(mask, iterations=num_erosions), dtype=int) for mask in masks]
    else:
        # clean up the mask
        masks = [binary_opening(mask) for mask in masks]

        # erode the mask without expanding holes
        if num_erosions:
            # use a bet masks to inform this, if available
            if bet_masks:
                bet_masks = [nib.load(bet_mask).get_fdata() for bet_mask in bet_masks]
                masks_filled = [np.array((masks[i] + bet_masks[i]) >= 1, dtype=int) for i in range(len(masks))]
            else:
                masks_filled = [fill_holes_morphological(mask) for mask in masks]

            # erode the mask without expanding holes
            masks_filled_ero = [np.array(binary_erosion(mask, iterations=num_erosions), dtype=int) for mask in masks_filled]
            
            holes = [np.array((masks_filled_ero[i] - masks[i]) == 1, int) for i in range(len(masks))]

            masks = [np.array((masks_filled_ero[i] - holes[i]) >= 1, dtype=int) for i in range(len(masks))]

    # determine filenames
    mask_filenames = [extend_fname(in_file, mask_suffix, out_dir=os.getcwd()) for in_file in in_files]

    for i in range(len(masks)):
        # set mask datatype to uint8
        all_niis[i].header.set_data_dtype(np.uint8)

        # add threshold information to nifti description
        threshold_description = f"Threshold = {round(thresholds[i], 3)}"
        description = str(all_niis[i].header['descrip'].astype(str)).strip()
        if description == "":
            all_niis[i].header['descrip'] = threshold_description
        else:
            all_niis[i].header['descrip'] = f"{description}; {threshold_description}"
        
        # save mask to file
        nib.save(
            nib.Nifti1Image(
                dataobj=masks[i],
                header=all_niis[i].header,
                affine=all_niis[i].affine
            ),
            mask_filenames[i]
        )

    return mask_filenames, thresholds

# The smoothing removes background noise and closes small holes
# A smaller threshold grows the mask
def fill_holes_smoothing(mask, sigma=[5,5,5], threshold=0.4):
    smoothed = gaussian_filter(mask * 1.0, sigma, truncate=2.0) # truncate reduces the kernel size: less precise but faster
    return np.array(smoothed > threshold, dtype=int)

# original morphological operation
def fill_holes_morphological(mask, fill_strength=0):
    filled_mask = mask.copy()
    for j in range(fill_strength):
        filled_mask = binary_dilation(filled_mask).astype(int)
    filled_mask = binary_fill_holes(filled_mask).astype(int)
    for j in range(fill_strength):
        filled_mask = binary_erosion(filled_mask).astype(int)
    return filled_mask

def fill_small_holes(mask):
    mask = mask.copy()
    kernel = np.array([
        [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
        [[1, 1, 1], [1, 0, 1], [1, 1, 1]],
        [[1, 1, 1], [1, 1, 1], [1, 1, 1]]
    ])
    weights = convolve(mask, kernel, mode='constant', cval=0.0)
    mask[weights >= 27 - 2] = 1
    return mask

class MaskingInputSpec(BaseInterfaceInputSpec):
    in_files = InputMultiPath(mandatory=True, exists=True)
    bet_masks = InputMultiPath(mandatory=False, exists=True)
    threshold = traits.Float(mandatory=False, default_value=None)
    fill_masks = traits.Bool(mandatory=False, default_value=True)
    mask_suffix = traits.String(mandatory=False, value="_mask")
    threshold_algorithm = traits.String(mandatory=False, value="otsu")
    threshold_algorithm_factor = traits.Float(mandatory=False, default_value=1.0)
    filling_algorithm = traits.String(mandatory=False, value='both')
    num_erosions = traits.Int(mandatory=False, default_value=0)


class MaskingOutputSpec(TraitedSpec):
    mask = OutputMultiPath(File(exists=False))
    threshold = traits.List()


class MaskingInterface(SimpleInterface):
    input_spec = MaskingInputSpec
    output_spec = MaskingOutputSpec

    def _run_interface(self, runtime):
        mask, threshold = threshold_masking(
            in_files=self.inputs.in_files,
            bet_masks=self.inputs.bet_masks,
            user_threshold=self.inputs.threshold,
            threshold_algorithm=self.inputs.threshold_algorithm,
            threshold_algorithm_factor=self.inputs.threshold_algorithm_factor,
            filling_algorithm=self.inputs.filling_algorithm,
            num_erosions=self.inputs.num_erosions,
            mask_suffix=f"_{self.inputs.mask_suffix}",
            fill_masks=self.inputs.fill_masks,
        )
        self._results['mask'] = mask
        self._results['threshold'] = threshold
        return runtime


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--in_files',
        nargs='+',
        required=True,
        type=str
    )

    parser.add_argument(
        '--threshold_value',
        type=float,
        default=None,
        help='Masking threshold for when --masking_algorithm is set to threshold. Values between 0 and 1'+
             'represent a percentage of the multi-echo input range. Values greater than 1 represent an '+
             'absolute threshold value. Lower values will result in larger masks. If no threshold is '+
             'provided, the --threshold_algorithm is used to select one automatically.'
    )

    parser.add_argument(
        '--threshold_algorithm',
        default='otsu',
        choices=['otsu', 'gaussian'],
        help='Algorithm used to select a threshold for threshold-based masking if --threshold_value is '+
             'left unspecified. The gaussian method is based on doi:10.1016/j.compbiomed.2012.01.004 '+
             'from Balan AGR. et al. The otsu method is based on doi:10.1109/TSMC.1979.4310076 from Otsu '+
             'et al.'
    )

    parser.add_argument(
        '--filling_algorithm',
        default='both',
        choices=['morphological', 'gaussian', 'both', 'none'],
        help='Algorithm used to fill holes for threshold-based masking. By default, a gaussian smoothing '+
             'operation is applied first prior to a morphological hole-filling operation. Note that gaussian '+
             'smoothing may fill some unwanted regions (e.g. connecting the skull and brain tissue), whereas '+
             'morphological hole-filling alone may fail to fill desired regions if they are not fully enclosed.'
    )

    parser.add_argument(
        '--num_erosions',
        default=1,
        type=int,
        help='Number of times to erode the mask.'
    )

    parser.add_argument(
        '--threshold_algorithm_factor',
        default=1.0,
        type=float,
        help='Factor to multiply the algorithmically-determined threshold by. Larger factors will create '+
             'smaller masks.'
    )


    args = parser.parse_args()
    mask_files = threshold_masking(
        in_files=args.in_files,
        user_threshold=args.threshold_value,
        threshold_algorithm=args.threshold_algorithm,
        threshold_algorithm_factor=args.threshold_algorithm_factor,
        filling_algorithm=args.filling_algorithm,
        num_erosions=args.num_erosions,
        fill_masks=args.filling_algorithm != 'none'
    )




#!/usr/bin/env python3
import nibabel as nib
import numpy as np
import os
import hashlib

from nipype.interfaces.base import SimpleInterface, BaseInterfaceInputSpec, TraitedSpec, traits, File
from qsmxt.scripts.qsmxt_functions import extend_fname

def frequency_to_normalized(frequency_path, B0, scale_factor=1, out_path=None):
    # use scale_factor=1e6 for microradians (needed for nextqsm)
    # load ΔB (Hz)
    frequency_nii = nib.load(frequency_path)
    ΔB = frequency_nii.get_fdata()
    
    # gyromagnetic ratio in Hz/T
    γ = 42.58*10**6

    # calculate normalized phase
    Φ_norm = (2*np.pi * ΔB) / (γ * B0) * scale_factor

    # save result
    out_path = out_path or extend_fname(frequency_path, f"_normalized", out_dir=os.getcwd())
    nib.save(img=nib.Nifti1Image(dataobj=Φ_norm, header=frequency_nii.header, affine=frequency_nii.affine), filename=out_path)
    return out_path


class FreqToNormalizedInputSpec(BaseInterfaceInputSpec):
    frequency = File(mandatory=True, exists=True)
    B0 = traits.Float(mandatory=True)
    scale_factor = traits.Float(mandatory=False, default_value=1)
    

class FreqToNormalizedOutputSpec(TraitedSpec):
    phase_normalized = File(mandatory=True, exists=True)


class FreqToNormalizedInterface(SimpleInterface):
    input_spec = FreqToNormalizedInputSpec
    output_spec = FreqToNormalizedOutputSpec

    def _run_interface(self, runtime):
        self._results['phase_normalized'] = frequency_to_normalized(
            frequency_path=self.inputs.frequency,
            B0=self.inputs.B0,
            scale_factor=self.inputs.scale_factor
        )
        return runtime


def frequency_to_phase(frequency_path, TE, wraps=False, out_path=None):
    # load ΔB (Hz)
    frequency_nii = nib.load(frequency_path)
    ΔB = frequency_nii.get_fdata()
    
    # phase accumulation at TE (rads)
    Φ_acc = 2*np.pi * ΔB * TE

    # add wraps if requested
    if wraps: Φ_acc = (Φ_acc + np.pi) % (2 * np.pi) - np.pi
    
    # save results
    out_path = out_path or extend_fname(frequency_path, f"_phase-TE{int(TE*1000):0>3}", out_dir=os.getcwd())
    nib.save(img=nib.Nifti1Image(dataobj=Φ_acc, header=frequency_nii.header, affine=frequency_nii.affine), filename=out_path)
    return out_path


class FreqToPhaseInputSpec(BaseInterfaceInputSpec):
    frequency = File(mandatory=True, exists=True)
    TE = traits.Float(mandatory=True)
    wraps = traits.Bool(default_value=False)
    

class FreqToPhaseOutputSpec(TraitedSpec):
    phase = File(mandatory=True, exists=True)


class FreqToPhaseInterface(SimpleInterface):
    input_spec = FreqToPhaseInputSpec
    output_spec = FreqToPhaseOutputSpec

    def _run_interface(self, runtime):
        self._results['phase'] = frequency_to_phase(
            frequency_path=self.inputs.frequency,
            TE=self.inputs.TE,
            wraps=self.inputs.wraps
        )
        return runtime


def phase_to_normalized(phase_path, B0, TE, scale_factor=1, out_path=None):
    # use scale_factor=1e6 for microradians (needed for nextqsm)
    # use scale_factor=1e6/(2*np.pi) for ??? (needed for rts, tv, etc.)

    # load phase
    phase_nii = nib.load(phase_path)
    Φ_acc = phase_nii.get_fdata()

    # gyromagnetic ratio in Hz/T
    γ = 42.58*10**6

    # calculate normalized phase
    Φ_norm = Φ_acc / (TE * γ * B0) * scale_factor

    # save result
    out_path = out_path or extend_fname(phase_path, f"_normalized", out_dir=os.getcwd())
    nib.save(img=nib.Nifti1Image(dataobj=Φ_norm, header=phase_nii.header, affine=phase_nii.affine), filename=out_path)

    return out_path


class PhaseToNormalizedInputSpec(BaseInterfaceInputSpec):
    phase = File(mandatory=True, exists=True)
    B0 = traits.Float(mandatory=False, default_value=3)
    TE = traits.Float(mandatory=True)
    scale_factor = traits.Float(mandatory=False, default_value=1)
    

class PhaseToNormalizedOutputSpec(TraitedSpec):
    phase_normalized = File(mandatory=True, exists=True)


class PhaseToNormalizedInterface(SimpleInterface):
    input_spec = PhaseToNormalizedInputSpec
    output_spec = PhaseToNormalizedOutputSpec

    def _run_interface(self, runtime):
        self._results['phase_normalized'] = phase_to_normalized(
            phase_path=self.inputs.phase,
            B0=self.inputs.B0,
            TE=self.inputs.TE,
            scale_factor=self.inputs.scale_factor
        )
        return runtime

def seed_from_filename(filename):
    hash_obj = hashlib.md5(filename.encode('utf-8'))
    hash_num = int(hash_obj.hexdigest(), 16)
    seed = hash_num % (2**32)  # Make seed 32-bit integer.
    return seed

def scale_to_pi(phase_path, phase_scaled_path=None):
    # load input phase
    phase_nii = nib.load(phase_path)
    Φ_acc_wrapped = phase_nii.get_fdata()

    # if a significant portion is exactly pi, replace with noise (GE data uses pi as the background)
    noise_added = False
    mask = np.logical_or((abs(abs(Φ_acc_wrapped) - np.pi)) < 1e-3, Φ_acc_wrapped == 0)
    if mask.sum() / mask.size >= 0.1:
        seed_value = seed_from_filename(phase_path)
        np.random.seed(seed_value)
        noise = np.random.uniform(-np.pi, np.pi, Φ_acc_wrapped.shape)
        Φ_acc_wrapped[mask] = noise[mask]
        noise_added = True
    
    # return the original if it is already scaled correctly
    if not noise_added and (np.round(np.min(Φ_acc_wrapped), 2)*-1) == np.round(np.max(Φ_acc_wrapped), 2) == 3.14:
        return phase_path
    
    # scale to -pi,+pi
    Φ_acc_wrapped_scaled = np.array(np.interp(Φ_acc_wrapped, (Φ_acc_wrapped.min(), Φ_acc_wrapped.max()), (-np.pi, +np.pi)), dtype=Φ_acc_wrapped.dtype)

    # save result
    phase_scaled_path = phase_scaled_path or extend_fname(phase_path, "_scaled", out_dir=os.getcwd())
    nib.save(nib.Nifti1Image(dataobj=Φ_acc_wrapped_scaled, header=phase_nii.header, affine=phase_nii.affine), phase_scaled_path)
    return phase_scaled_path


class ScalePhaseInputSpec(BaseInterfaceInputSpec):
    phase = File(mandatory=True, exists=True)
    

class ScalePhaseOutputSpec(TraitedSpec):
    phase = File(mandatory=True, exists=True)


class ScalePhaseInterface(SimpleInterface):
    input_spec = ScalePhaseInputSpec
    output_spec = ScalePhaseOutputSpec

    def _run_interface(self, runtime):
        self._results['phase'] = scale_to_pi(self.inputs.phase)
        return runtime


